

from typing import Optional
from langport.model.executor.base import BaseModelExecutor
# from langport.workers.embedding_worker import EmbeddingModelWorker


class EmbeddingExecutor(BaseModelExecutor):
    def __init__(
        self,
        model_path: str,
        model_name: str,
        device: str,
        num_gpus: int,
        max_gpu_memory: Optional[str],
        load_8bit: bool,
        cpu_offloading: bool,
    ) -> None:
        super(EmbeddingExecutor, self).__init__(
            model_path=model_path,
            model_name=model_name,
            device=device,
            num_gpus=num_gpus,
            max_gpu_memory=max_gpu_memory,
            load_8bit=load_8bit,
            cpu_offloading=cpu_offloading
        )
        self.adapter = None
        self.model = None
        self.tokenizer = None
    
    def inference(self, worker: "EmbeddingModelWorker"):
        raise NotImplementedError("EmbeddingExecutor inference not implemented")
        
    


from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
)

from langport.data.conversation import ConversationHistory, SeparatorStyle
from langport.data.conversation.conversation_settings import get_conv_settings
from langport.model.model_adapter import BaseAdapter

class PhoenixAdapter(BaseAdapter):
    """The model adapter for FreedomIntelligence/phoenix-inst-chat-7b"""

    def match(self, model_path: str):
        return "phoenix" in model_path

    def get_default_conv_template(self, model_path: str) -> ConversationHistory:
        settings = get_conv_settings("phoenix")
        return ConversationHistory(
            system="A chat between a curious human and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the human's questions.\n\n",
            messages=(),
            offset=0,
            settings=settings,
        )



from langport.data.conversation import (
    ConversationSettings,
    SeparatorStyle,
)

# OpenAssistant Pythia default template
oasst_pythia = ConversationSettings(
    name="oasst_pythia",
    roles=("<|prompter|>", "<|assistant|>"),
    sep_style=SeparatorStyle.NO_COLON_SINGLE,
    sep="<|endoftext|>",
)



from typing import Optional

from langport.model.adapters.dolly_v2 import DollyV2Adapter
from langport.model.adapters.rwkv import RwkvAdapter
from langport.model.adapters.t5 import T5Adapter
from langport.model.adapters.text2vec import BertAdapter
from langport.model.executor.base import LocalModelExecutor

import torch

from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    AutoModelForSeq2SeqLM,
    T5Tokenizer,
    BertTokenizer,
    BertModel,
)


import math
from typing import Optional
import warnings
import psutil

import torch
from langport.model.compression import load_compress_model
from langport.model.executor.base import BaseModelExecutor
from langport.model.model_adapter import get_model_adapter, raise_warning_for_incompatible_cpu_offloading_configuration
from langport.model.monkey_patch_non_inplace import replace_llama_attn_with_non_inplace_operations
from langport.utils import get_gpu_memory



class HuggingfaceExecutor(LocalModelExecutor):
    def __init__(
        self,
        model_name: str,
        model_path: str,
        device: str,
        num_gpus: int,
        max_gpu_memory: Optional[str],
        load_8bit: bool,
        cpu_offloading: bool,
        deepspeed: bool = False,
    ) -> None:
        super(HuggingfaceExecutor, self).__init__(
            model_name=model_name,
            model_path=model_path,
            device=device,
            num_gpus=num_gpus,
            max_gpu_memory=max_gpu_memory,
            load_8bit=load_8bit,
            cpu_offloading=cpu_offloading
        )
    
      
    def _load_hf_model(self, adapter, model_path: str, from_pretrained_kwargs: dict):
        if isinstance(adapter, DollyV2Adapter):
            tokenizer = AutoTokenizer.from_pretrained(model_path, use_fast=False)
            model = AutoModelForCausalLM.from_pretrained(
                model_path, low_cpu_mem_usage=True, **from_pretrained_kwargs
            )
            # 50277 means "### End"
            tokenizer.eos_token_id = 50277
        elif isinstance(adapter, RwkvAdapter):
            from langport.model.models.rwkv_model import RwkvModel
            model = RwkvModel(model_path)
            tokenizer = AutoTokenizer.from_pretrained(
                "EleutherAI/pythia-160m", use_fast=True
            )
        elif isinstance(adapter, T5Adapter):
            tokenizer = T5Tokenizer.from_pretrained(model_path, use_fast=False)
            model = AutoModelForSeq2SeqLM.from_pretrained(
                model_path, low_cpu_mem_usage=True, **from_pretrained_kwargs
            )
        elif isinstance(adapter, BertAdapter):
            tokenizer = BertTokenizer.from_pretrained(model_path)
            model = BertModel.from_pretrained(model_path, **from_pretrained_kwargs)

        else:
            tokenizer = AutoTokenizer.from_pretrained(model_path, use_fast=False)
            model = AutoModelForCausalLM.from_pretrained(
                model_path, low_cpu_mem_usage=True, **from_pretrained_kwargs
            )

        return model, tokenizer

    def load_model(
        self,
        model_path: str,
        device: str,
        num_gpus: int,
        max_gpu_memory: Optional[str] = None,
        load_8bit: bool = False,
        cpu_offloading: bool = False,
        deepspeed: bool = False,
        trust_remote_code: bool = False,
        debug: bool = False,
    ):
        """Load a model from Hugging Face."""
        adapter = get_model_adapter(model_path)

        # Handle device mapping
        cpu_offloading = raise_warning_for_incompatible_cpu_offloading_configuration(
            device, load_8bit, cpu_offloading
        )
        if device == "cpu":
            kwargs = {"torch_dtype": torch.float32}
        elif device == "cuda":
            kwargs = {"torch_dtype": torch.float16}
            if num_gpus != 1:
                kwargs["device_map"] = "auto"
                if max_gpu_memory is None:
                    kwargs[
                        "device_map"
                    ] = "sequential"  # This is important for not the same VRAM sizes
                    available_gpu_memory = get_gpu_memory(num_gpus)
                    kwargs["max_memory"] = {
                        i: str(int(available_gpu_memory[i] * 0.85)) + "GiB"
                        for i in range(num_gpus)
                    }
                else:
                    kwargs["max_memory"] = {i: max_gpu_memory for i in range(num_gpus)}
        elif device == "mps":
            kwargs = {"torch_dtype": torch.float16}
            # Avoid bugs in mps backend by not using in-place operations.
            replace_llama_attn_with_non_inplace_operations()
        else:
            raise ValueError(f"Invalid device: {device}")

        kwargs["trust_remote_code"] = trust_remote_code

        if cpu_offloading:
            # raises an error on incompatible platforms
            from transformers import BitsAndBytesConfig

            if "max_memory" in kwargs:
                kwargs["max_memory"]["cpu"] = (
                    str(math.floor(psutil.virtual_memory().available / 2**20)) + "Mib"
                )
            kwargs["quantization_config"] = BitsAndBytesConfig(
                load_in_8bit_fp32_cpu_offload=cpu_offloading
            )
            kwargs["load_in_8bit"] = load_8bit
            # Load model
            model, tokenizer = self._load_hf_model(adapter, model_path, kwargs)
        elif load_8bit:
            if num_gpus != 1:
                warnings.warn(
                    "8-bit quantization is not supported for multi-gpu inference."
                )
            else:
                model, tokenizer = load_compress_model(
                    model_path=model_path, device=device, torch_dtype=kwargs["torch_dtype"]
                )
                # return adapter, model, tokenizer
        else:
            # Load model
            model, tokenizer = self._load_hf_model(adapter, model_path, kwargs)

        if deepspeed:
            from transformers.deepspeed import HfDeepSpeedConfig
            import deepspeed

            dtype = torch.float16
            if load_8bit:
                dtype = torch.int8
            config = {
                # "mp_size": 1,        # Number of GPU
                "dtype": dtype, # dtype of the weights (fp16)
                "replace_method": "auto", # Lets DS autmatically identify the layer to replace
                "replace_with_kernel_inject": True, # replace the model with the kernel injector
                "enable_cuda_graph": True,
                "tensor_parallel": {
                    "enabled": True,
                    "tp_size": 1,
                },
            }
            ds_engine = deepspeed.init_inference(model=model, config=config)
            model = ds_engine.module
        else:
            if (device == "cuda" and num_gpus == 1 and not cpu_offloading) or device == "mps":
                model.to(device)

        if debug:
            print(model)

        return adapter, model, tokenizer




def add_model_args(parser):
    parser.add_argument(
        "--model-path",
        type=str,
        default="lmsys/fastchat-t5-3b-v1.0",
        help="The path to the weights. This can be a local folder or a Hugging Face repo ID.",
    )
    parser.add_argument(
        "--device",
        type=str,
        choices=["cpu", "cuda", "mps"],
        default="cuda",
        help="The device type",
    )
    parser.add_argument(
        "--gpus",
        type=str,
        default=None,
        help="A single GPU like 1 or multiple GPUs like 0,2",
    )
    parser.add_argument("--num-gpus", type=int, default=1)
    parser.add_argument(
        "--max-gpu-memory",
        type=str,
        help="The maximum memory per gpu. Use a string like '13Gib'",
    )
    parser.add_argument(
        "--load-8bit", action="store_true", help="Use 8-bit quantization"
    )
    parser.add_argument(
        "--cpu-offloading",
        action="store_true",
        help="Only when using 8-bit quantization: Offload excess weights to the CPU that don't fit on the GPU",
    )
    parser.add_argument(
        "--deepspeed", action="store_true", help="Use deepspeed"
    )
    parser.add_argument(
        "--trust-remote-code", action="store_true", help="Trust remote code"
    )


from typing import Any, Iterable, List, Optional, Union
from langport.model.executor.generation import BaseStreamer

from langport.model.executor.huggingface import HuggingfaceExecutor

from langport.protocol.worker_protocol import (
    BaseWorkerResult,
    GenerationTask,
    GenerationWorkerResult,
    UsageInfo,
)

import torch

from transformers import PreTrainedModel, PreTrainedTokenizerBase
from transformers.generation.logits_process import (
    LogitsProcessor,
    LogitsProcessorList,
    TemperatureLogitsWarper,
    RepetitionPenaltyLogitsProcessor,
    TopPLogitsWarper,
    TopKLogitsWarper,
)
from langport.workers.generation_worker import GenerationModelWorker

from cachetools import LRUCache, TTLCache
from asyncache import cached

from typing import Optional

import torch

@cached(LRUCache(maxsize=64))
def prepare_logits_processor(
    temperature: float, repetition_penalty: float, top_p: float, top_k: int
) -> LogitsProcessorList:
    processor_list = LogitsProcessorList()
    # TemperatureLogitsWarper doesn't accept 0.0, 1.0 makes it a no-op so we skip two cases.
    if temperature >= 1e-5 and temperature != 1.0:
        processor_list.append(TemperatureLogitsWarper(temperature))
    if repetition_penalty > 1.0:
        processor_list.append(RepetitionPenaltyLogitsProcessor(repetition_penalty))
    if 1e-8 <= top_p < 1.0:
        processor_list.append(TopPLogitsWarper(top_p))
    if top_k > 0:
        processor_list.append(TopKLogitsWarper(top_k))
    return processor_list


class BatchingTask:
    def __init__(self, tasks: List[GenerationTask], tokenizer: PreTrainedTokenizerBase, device: str, is_encoder_decoder: bool) -> None:
        self.batch_size = len(tasks)
        if self.batch_size == 0:
            return
        
        self.tokenizer = tokenizer
        self.device = device
        self.is_encoder_decoder = is_encoder_decoder
        
        # collect params
        self.tasks = tasks
        self.prompts_ids = [self.tokenizer(task.prompt, return_tensors="pt").input_ids.squeeze(0) for task in tasks]
        self.prompts_ids_length = [len(i) for i in self.prompts_ids]
        self.min_prompts_length = min(self.prompts_ids_length)
        self.max_prompts_length = max(self.prompts_ids_length)
        self.max_tokens = [task.max_tokens for task in tasks]
        if self.tokenizer._pad_token is None:
            self.pad_fill_id = self.tokenizer.eos_token_id
        else:
            self.pad_fill_id = self.tokenizer.pad_token_id

        # init logits_processor
        self.logits_processor_list = []
        for task in tasks:
            logits_processor = prepare_logits_processor(
                task.temperature, task.repetition_penalty, task.top_p, task.top_k
            )
            self.logits_processor_list.append(logits_processor)
        
        # variables used in the streaming process
        self.batch_tokens_cache: List[List[int]] = [[] for i in range(self.batch_size)]
        self.stop = [False for i in range(self.batch_size)]

    def __len__(self):
        return self.batch_size
    
    def __call__(self, return_attention_mask=True) -> Any:
        ids = [torch.cat((self.prompts_ids[i], torch.LongTensor(self.batch_tokens_cache[i])))
                         for i in range(self.batch_size)]
        length = [len(i) for i in ids]
        # padding to max(length)
        full_input_ids = torch.full(
            (self.batch_size, max(length)), self.pad_fill_id, dtype=torch.long, device=self.device
        )
        for i in range(self.batch_size):
            if self.is_encoder_decoder:
                full_input_ids[i, :length[i]] = ids[i]
            else:
                full_input_ids[i, -length[i]:] = ids[i] # padding side left
        if not return_attention_mask:
            return full_input_ids
        return full_input_ids, self._gen_attention_mask(length)
    
    def _gen_attention_mask(self, length: List[int]) -> torch.Tensor:
        mask = torch.full(
            (self.batch_size, max(length)), 0, dtype=torch.long, device=self.device
        )
        if self.is_encoder_decoder:
            for i in range(self.batch_size):
                mask[i, :length[i]] = 1
        else:
            for i in range(self.batch_size):
                mask[i, -length[i]:] = 1
        return mask
    
    def _check_idx(self, idx:int):
        if idx > self.batch_size:
            raise ValueError("Invalid batch index")
    
    def _check_batch_size(self, lenable):
        if len(lenable) != self.batch_size:
            raise ValueError("Different batch size")
    
    def get_prompt_ids(self, idx:int) -> List[int]:
        self._check_idx(idx)
        return self.prompts_ids[idx]
    
    def get_prompt_length(self, idx:int) -> int:
        return len(self.get_prompt_ids(idx))
    
    def get_logits_processor_list(self, idx:int) -> LogitsProcessorList:
        self._check_idx(idx)
        return self.logits_processor_list[idx]
    
    def get_generated_ids(self, idx:int) -> List[int]:
        self._check_idx(idx)
        return self.batch_tokens_cache[idx]
    
    def get_generated_length(self, idx:int) -> int:
        return len(self.get_generated_ids(idx))
    
    def update_new_token(self, batch_token: List[int]):
        self._check_batch_size(batch_token)
        for i, token in enumerate(batch_token):
            if self.is_stop(i):
                continue
            self.batch_tokens_cache[i].append(token)

            # auto check stop
            if token == self.tokenizer.eos_token_id:
                self.set_stop(i)
            if self.tasks[i].stop_token_ids is not None and token in self.tasks[i].stop_token_ids:
                self.set_stop(i)
            if self.get_prompt_length(i) + self.get_generated_length(i) >= self.max_tokens[i]:
                self.set_stop(i)
                
    def set_stop(self, idx:int):
        self._check_idx(idx)
        self.stop[idx] = True
    
    def is_stop(self, idx:int):
        self._check_idx(idx)
        return self.stop[idx]


class GenerationModel:
    def __init__(self, model: PreTrainedModel) -> None:
        self.model = model
    
    @torch.inference_mode()
    def generate(self, inputs: BatchingTask, 
                 max_new_tokens: int,
                 streamer: Optional[BaseStreamer]=None) -> torch.Tensor:

        if inputs.batch_size == 0:
            return

        full_input_ids, attention_mask = inputs()
        if self.model.config.is_encoder_decoder:
            encoder_outputs = self.model.encoder(input_ids=full_input_ids, attention_mask=attention_mask)
            decoder_input_ids_list = [torch.full(
                (inputs.batch_size, 1),
                self.model.generation_config.decoder_start_token_id,
                dtype=torch.long,
                device=self.model.device,
            )]
        else:
            encoder_outputs = None
            decoder_input_ids_list = [full_input_ids]
        past_key_values = None

        # step by step
        for step in range(max_new_tokens):
            # inference a step
            if len(decoder_input_ids_list) > 1:
                decoder_input_ids = torch.stack(decoder_input_ids_list, dim=1)
            elif len(decoder_input_ids_list) == 1:
                decoder_input_ids = decoder_input_ids_list[0]
            else:
                raise Exception("decoder_input_ids_list length is 0")
            if self.model.config.is_encoder_decoder:
                out = self.model.decoder(
                    input_ids=decoder_input_ids,
                    use_cache=self.model.generation_config.use_cache,
                    encoder_outputs=encoder_outputs,
                    past_key_values=past_key_values,
                )
            else:
                out = self.model(
                    input_ids=decoder_input_ids,
                    use_cache=self.model.generation_config.use_cache,
                    past_key_values=past_key_values,
                )
            logits = out.logits
            past_key_values = out.past_key_values
            decoder_input_ids_list = []

            new_ids = []

            for i, task in enumerate(inputs.tasks):
                if inputs.is_stop(i):
                    new_ids.append(inputs.pad_fill_id)
                    continue
                each_logits = logits[i, -1, :].unsqueeze(0)

                logits_processor = inputs.get_logits_processor_list(i)
                if logits_processor:
                    if task.repetition_penalty > 1.0:
                        tmp_output_ids = decoder_input_ids[i, :].unsqueeze(0)
                    else:
                        tmp_output_ids = None
                    last_token_logits = logits_processor(tmp_output_ids, each_logits)[0]
                else:
                    last_token_logits = each_logits

                if self.model.device.type == "mps":
                    # Switch to CPU by avoiding some bugs in mps backend.
                    last_token_logits = last_token_logits.float().to("cpu")

                if task.temperature < 1e-5 or task.top_p < 1e-8:  # greedy
                    token = int(torch.argmax(last_token_logits))
                else:
                    probs = torch.softmax(last_token_logits, dim=-1)
                    token = int(torch.multinomial(probs, num_samples=1))
                new_ids.append(token)
            
            # update state
            inputs.update_new_token(new_ids)
            if streamer:
                streamer.put(new_ids)

            if self.model.config.is_encoder_decoder or self.model.generation_config.use_cache:
                # print("use cache!")
                new_ids_tensor = torch.tensor(
                    new_ids, dtype=torch.long, device=decoder_input_ids.device).unsqueeze(1)
                decoder_input_ids_list = [new_ids_tensor]
            else:
                decoder_input_ids = inputs(return_attention_mask=False)
                decoder_input_ids_list = [decoder_input_ids]

            if all(inputs.stop):
                break

        if streamer:
            streamer.end()

        del past_key_values

class GenerationWorkerStreamer(BaseStreamer):
    def __init__(self,
                 task_batch: BatchingTask,
                 tokenizer: PreTrainedTokenizerBase,
                 worker: "GenerationModelWorker") -> None:
        self.task_batch = task_batch
        self.tokenizer = tokenizer
        self.worker = worker
        self.stream_interval = worker.stream_interval

        self.done = [False for i in range(task_batch.batch_size)]

    def put(self, value):
        for i in range(self.task_batch.batch_size):
            generated_len = self.task_batch.get_generated_length(i)
            if (self.done[i] or generated_len % self.stream_interval != 0) and self.done[i]==self.task_batch.is_stop(i):
                continue
            task = self.task_batch.tasks[i]
            text = self.tokenizer.decode(self.task_batch.get_generated_ids(i), skip_special_tokens=True)
            stop_pos = stop_by_stopwords(text, 0, task.stop)
            if stop_pos != -1:
                self.task_batch.set_stop(i)
                text = text[:stop_pos]
            prompt_len = self.task_batch.get_prompt_length(i)

            if self.task_batch.is_stop(i):
                if prompt_len + generated_len >= self.task_batch.max_tokens[i]:
                    finish_reason = "length"
                else:
                    finish_reason = "stop"
                self.worker.push_task_result(task.task_id,
                    GenerationWorkerResult(
                        task_id=task.task_id,
                        type="finish",
                        text=text,
                        usage=UsageInfo(
                            prompt_tokens=prompt_len,
                            total_tokens=prompt_len + generated_len,
                            completion_tokens=generated_len,
                        ),
                        finish_reason=finish_reason,
                    )
                )
                self.worker.push_task_result(task.task_id,
                    BaseWorkerResult(task_id=task.task_id, type="done")
                )
                self.done[i] = True
            else:
                self.worker.push_task_result(task.task_id,
                    GenerationWorkerResult(
                        task_id=task.task_id,
                        type="data",
                        text=text,
                        usage=UsageInfo(
                            prompt_tokens=prompt_len,
                            total_tokens=prompt_len + generated_len,
                            completion_tokens=generated_len,
                        ),
                        finish_reason=None,
                    )
                )

    def end(self):
        pass
            

def stop_by_stopwords(
    output: str, rfind_start: int, stop: Optional[Union[str, List[str]]]
) -> int:
    if stop is not None:
        if isinstance(stop, str):
            pos = output.rfind(stop, rfind_start)
            if pos != -1:
                return pos
        elif isinstance(stop, Iterable):
            for each_stop in stop:
                pos = output.rfind(each_stop, rfind_start)
                if pos != -1:
                    return pos
        else:
            raise ValueError("Invalid stop field type.")
    return -1

class HuggingfaceGenerationExecutor(HuggingfaceExecutor):
    def __init__(
        self,
        model_name: str,
        model_path: str,
        device: str,
        num_gpus: int,
        max_gpu_memory: Optional[str],
        load_8bit: bool,
        cpu_offloading: bool,
        deepspeed: bool = False,
        trust_remote_code: bool = False
    ) -> None:
        super(HuggingfaceGenerationExecutor, self).__init__(
            model_name=model_name,
            model_path=model_path,
            device=device,
            num_gpus=num_gpus,
            max_gpu_memory=max_gpu_memory,
            load_8bit=load_8bit,
            cpu_offloading=cpu_offloading
        )
        self.adapter = None
        self.model = None
        self.tokenizer = None
        self.adapter, self.model, self.tokenizer = self.load_model(
            model_path, device, num_gpus, max_gpu_memory, load_8bit, cpu_offloading, deepspeed, trust_remote_code
        )

        # self.model = torch.compile(self.model)

        if hasattr(self.model.config, "max_sequence_length"):
            self._context_len = self.model.config.max_sequence_length
        elif hasattr(self.model.config, "max_position_embeddings"):
            self._context_len = self.model.config.max_position_embeddings
        else:
            self._context_len = 2048

    @property
    def context_length(self) -> int:
        return self._context_len
    
    def tokenize(self, text: str) -> List[int]:
        input_ids = self.tokenizer(text).input_ids
        return input_ids
    
    def inference(self, worker: "GenerationModelWorker"):
        if not worker.online:
            return

        tasks = worker.fetch_tasks()

        # batch inference
        inputs = BatchingTask(tasks, self.tokenizer, self.device, self.model.config.is_encoder_decoder)
        if inputs.batch_size == 0:
            return
        streamer = GenerationWorkerStreamer(inputs, self.tokenizer, worker)
        model = GenerationModel(self.model)
        max_new_tokens = max(inputs.max_tokens)
        model.generate(inputs, max_new_tokens, streamer)
  


from transformers import AutoModelForCausalLM, AutoTokenizer
from langport.model.model_adapter import BaseAdapter

class StarCoderAdapter(BaseAdapter):
    """The model adapter for starcoder models from bigcode-project."""

    def match(self, model_path: str):
        return "starcoder" in model_path



import asyncio

import argparse
import asyncio
import json
import logging

import os
import random
import traceback
from typing import Generator, Optional, Union, Dict, List, Any

import fastapi
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import StreamingResponse, JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware, DispatchFunction
import httpx
import numpy as np
import shortuuid
from starlette.types import ASGIApp
from tenacity import retry, stop_after_attempt
import uvicorn
from pydantic import BaseSettings

from langport.constants import WORKER_API_TIMEOUT, ErrorCode
from langport.model.model_adapter import get_conversation_template
from fastapi.exceptions import RequestValidationError
from langport.protocol.openai_api_protocol import (
    ChatCompletionRequest,
    ChatCompletionResponse,
    ChatCompletionResponseStreamChoice,
    ChatCompletionStreamResponse,
    ChatMessage,
    ChatCompletionResponseChoice,
    CompletionRequest,
    CompletionResponse,
    CompletionResponseChoice,
    DeltaMessage,
    CompletionResponseStreamChoice,
    CompletionStreamResponse,
    EmbeddingsData,
    EmbeddingsRequest,
    EmbeddingsResponse,
    ErrorResponse,
    ModelCard,
    ModelList,
    ModelPermission,
    UsageInfo,
)
from langport.protocol.worker_protocol import (
    BaseWorkerResult,
    EmbeddingWorkerResult,
    GenerationWorkerResult,
    WorkerAddressRequest,
    WorkerAddressResponse,
)
from langport.core.dispatch import DispatchMethod
from langport.routers.gateway.common import LANGPORT_HEADER, AppSettings, _get_worker_address, _list_models, check_model, check_requests, create_error_response


def get_gen_params(
    model_name: str,
    messages: Union[str, List[Dict[str, str]]],
    *,
    temperature: float,
    top_p: float,
    max_tokens: Optional[int],
    echo: Optional[bool],
    stream: Optional[bool],
    stop: Optional[Union[str, List[str]]],
) -> Dict[str, Any]:
    is_chatglm = "chatglm" in model_name.lower()
    conv = get_conversation_template(model_name)

    if isinstance(messages, str):
        prompt = messages
    else:
        for message in messages:
            msg_role = message["role"]
            if msg_role == "system":
                conv.system = message["content"]
            elif msg_role == "user":
                conv.append_message(conv.settings.roles[0], message["content"])
            elif msg_role == "assistant":
                conv.append_message(conv.settings.roles[1], message["content"])
            else:
                raise ValueError(f"Unknown role: {msg_role}")

        # Add a blank message for the assistant.
        conv.append_message(conv.settings.roles[1], None)

        if is_chatglm:
            prompt = conv.messages[conv.offset :]
        else:
            prompt = conv.get_prompt()

    if max_tokens is None:
        max_tokens = 512

    gen_params = {
        "model": model_name,
        "prompt": prompt,
        "temperature": temperature,
        "top_p": top_p,
        "max_tokens": max_tokens,
        "echo": echo,
        "stream": stream,
    }

    if stop is None:
        gen_params.update(
            {"stop": conv.settings.stop_str, "stop_token_ids": conv.settings.stop_token_ids}
        )
    else:
        gen_params.update({"stop": stop})

    return gen_params

async def api_models(app_settings: AppSettings):
    async with httpx.AsyncClient() as client:
        models = await _list_models(app_settings, None, client)
    models.sort()
    # TODO: return real model permission details
    model_cards = []
    for m in models:
        model_cards.append(ModelCard(id=m, root=m, permission=[ModelPermission()]))
    return ModelList(data=model_cards)

async def generate_completion_stream_generator(app_settings: AppSettings, payload: Dict[str, Any], n: int):
    model_name = payload["model"]
    id = f"cmpl-{shortuuid.random()}"
    finish_stream_events = []

    for i in range(n):
        previous_text = ""
        async for content in generate_completion_stream(app_settings, "/completion_stream", payload):
            if content.error_code != ErrorCode.OK:
                yield f"data: {json.dumps(content.dict(), ensure_ascii=False)}\n\n"
                yield "data: [DONE]\n\n"
                return
            decoded_unicode = content.text.replace("\ufffd", "")
            delta_text = decoded_unicode[len(previous_text) :]
            previous_text = decoded_unicode

            choice_data = CompletionResponseStreamChoice(
                index=i,
                text=delta_text,
                logprobs=content.logprobs,
                finish_reason=content.finish_reason,
            )
            chunk = CompletionStreamResponse(
                id=id, object="text_completion", choices=[choice_data], model=model_name
            )
            if len(delta_text) == 0:
                if content.finish_reason is not None:
                    finish_stream_events.append(chunk)
                continue
            yield f"data: {chunk.json(exclude_unset=True, ensure_ascii=False)}\n\n"

    # There is not "content" field in the last delta message, so exclude_none to exclude field "content".
    for finish_chunk in finish_stream_events:
        yield f"data: {finish_chunk.json(exclude_unset=True, ensure_ascii=False)}\n\n"
    yield "data: [DONE]\n\n"


async def generate_completion_stream(app_settings: AppSettings, url: str, payload: Dict[str, Any]) -> Generator[GenerationWorkerResult, Any, None]:
    async with httpx.AsyncClient() as client:
        try:
            worker_addr = await _get_worker_address(app_settings, payload["model"], "generation", client, DispatchMethod.LOTTERY)
        except:
            yield BaseWorkerResult(
                type="error",
                message=f"No available worker running {payload['model']} for generation",
                error_code=ErrorCode.INVALID_MODEL
            )
            return

        delimiter = b"\0"
        try:
            async with client.stream(
                "POST",
                worker_addr + url,
                headers=LANGPORT_HEADER,
                json=payload,
                timeout=WORKER_API_TIMEOUT,
            ) as response:
                async for raw_chunk in response.aiter_raw():
                    for chunk in raw_chunk.split(delimiter):
                        if not chunk:
                            continue
                        try:
                            data = json.loads(chunk.decode())
                        except json.JSONDecodeError:
                            yield BaseWorkerResult(
                                type="error",
                                message=chunk.decode(),
                                error_code=ErrorCode.ENGINE_OVERLOADED
                            )
                            break
                        if data["type"] == "error":
                            yield BaseWorkerResult.parse_obj(data)
                            break
                        yield GenerationWorkerResult.parse_obj(data)
        except httpx.ReadTimeout:
            yield BaseWorkerResult(
                type="error",
                message="Server is overloading",
                error_code=ErrorCode.ENGINE_OVERLOADED
            )
        

async def single_completions_non_stream(app_settings: AppSettings, payload: Dict[str, Any]) -> Optional[GenerationWorkerResult]:
    ret = None
    async for content in generate_completion_stream(app_settings, "/completion_stream", payload):
        ret = content
    return ret

async def chat_completion_stream_generator(
    app_settings: AppSettings, payload: Dict[str, Any], n: int
) -> Generator[str, Any, None]:
    """
    Event stream format:
    https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#event_stream_format
    """
    id = f"chatcmpl-{shortuuid.random()}"
    finish_stream_events = []
    for i in range(n):
        # First chunk with role
        choice_data = ChatCompletionResponseStreamChoice(
            index=i,
            delta=DeltaMessage(role="assistant"),
            finish_reason=None,
        )
        chunk = ChatCompletionStreamResponse(
            id=id, choices=[choice_data], model=payload["model"]
        )
        yield f"data: {chunk.json(exclude_unset=True, ensure_ascii=False)}\n\n"

        previous_text = ""
        async for content in generate_completion_stream(app_settings, "/chat_stream", payload):
            if content.error_code != ErrorCode.OK:
                yield f"data: {json.dumps(content.dict(), ensure_ascii=False)}\n\n"
                yield "data: [DONE]\n\n"
                return
            decoded_unicode = content.text.replace("\ufffd", "")
            delta_text = decoded_unicode[len(previous_text) :]
            previous_text = decoded_unicode

            if len(delta_text) == 0:
                delta_text = None
            choice_data = ChatCompletionResponseStreamChoice(
                index=i,
                delta=DeltaMessage(content=delta_text),
                finish_reason=content.finish_reason,
            )
            chunk = ChatCompletionStreamResponse(
                id=id, choices=[choice_data], model=payload["model"]
            )
            if delta_text is None:
                if content.finish_reason is not None:
                    finish_stream_events.append(chunk)
                continue
            yield f"data: {chunk.json(exclude_unset=True, ensure_ascii=False)}\n\n"
    # There is not "content" field in the last delta message, so exclude_none to exclude field "content".
    for finish_chunk in finish_stream_events:
        yield f"data: {finish_chunk.json(exclude_none=True, ensure_ascii=False)}\n\n"
    yield "data: [DONE]\n\n"


async def single_chat_completions_non_stream(
    app_settings: AppSettings, payload: Dict[str, Any]
) -> Optional[GenerationWorkerResult]:
    ret = None
    async for content in generate_completion_stream(app_settings, "/chat_stream", payload):
        ret = content
    return ret


async def get_embedding(app_settings: AppSettings, payload: Dict[str, Any]) -> EmbeddingWorkerResult:
    model_name = payload["model"]
    async with httpx.AsyncClient() as client:
        worker_addr = await _get_worker_address(app_settings, model_name, "embedding", client, DispatchMethod.LOTTERY)

        response = await client.post(
            worker_addr + "/embeddings",
            headers=LANGPORT_HEADER,
            json=payload,
            timeout=WORKER_API_TIMEOUT,
        )
        if response.json()["type"] == "error":
            error_message = BaseWorkerResult.parse_obj(response.json())
            return error_message
        return EmbeddingWorkerResult.parse_obj(response.json())


async def completions_stream(app_settings: AppSettings, payload: Dict[str, Any], request: CompletionRequest):
    generator = generate_completion_stream_generator(app_settings, payload, request.n)
    return StreamingResponse(generator, media_type="text/event-stream")

async def completions_non_stream(app_settings: AppSettings, payload: Dict[str, Any], request: CompletionRequest):
    completions = []
    for i in range(request.n):
        content = asyncio.create_task(single_completions_non_stream(app_settings, payload))
        completions.append(content)

    choices = []
    usage = UsageInfo()
    for i, content_task in enumerate(completions):
        content = await content_task
        if content.error_code != ErrorCode.OK:
            return create_error_response(content.error_code, content.message)
        choices.append(
            CompletionResponseChoice(
                index=i,
                text=content.text,
                logprobs=content.logprobs,
                finish_reason=content.finish_reason,
            )
        )
        task_usage = UsageInfo.parse_obj(content.usage)
        for usage_key, usage_value in task_usage.dict().items():
            setattr(usage, usage_key, getattr(usage, usage_key) + usage_value)

    return CompletionResponse(
        model=request.model, choices=choices, usage=UsageInfo.parse_obj(usage)
    )

async def chat_completions_stream(app_settings: AppSettings, payload: Dict[str, Any], request: ChatCompletionRequest):
    generator = chat_completion_stream_generator(
        app_settings, payload, request.n
    )
    return StreamingResponse(generator, media_type="text/event-stream")


async def chat_completions_non_stream(app_settings: AppSettings, payload: Dict[str, Any], request: ChatCompletionRequest):
    choices = []
    chat_completions = []
    for i in range(request.n):
        content = asyncio.create_task(single_chat_completions_non_stream(app_settings, payload))
        chat_completions.append(content)

    usage = UsageInfo()
    for i, content_task in enumerate(chat_completions):
        content = await content_task
        if content is None:
            return create_error_response(ErrorCode.INTERNAL_ERROR, "Server internal error")
        if content.error_code != ErrorCode.OK:
            return create_error_response(content.error_code, content.message)
        choices.append(
            ChatCompletionResponseChoice(
                index=i,
                message=ChatMessage(role="assistant", content=content.text),
                finish_reason=content.finish_reason,
            )
        )
        if content.usage is None:
            continue
        task_usage = UsageInfo.parse_obj(content.usage)
        for usage_key, usage_value in task_usage.dict().items():
            setattr(usage, usage_key, getattr(usage, usage_key) + usage_value)

    return ChatCompletionResponse(model=request.model, choices=choices, usage=usage)


async def api_completions(app_settings: AppSettings, request: CompletionRequest):
    error_check_ret = await check_model(app_settings, request, "generation", request.model)
    if error_check_ret is not None:
        return error_check_ret
    error_check_ret = check_requests(request)
    if error_check_ret is not None:
        return error_check_ret

    payload = get_gen_params(
        request.model,
        request.prompt,
        temperature=request.temperature,
        top_p=request.top_p,
        max_tokens=request.max_tokens,
        echo=request.echo,
        stream=request.stream,
        stop=request.stop,
    )

    if request.stream:
        return await completions_stream(app_settings, payload, request)
    else:
        return await completions_non_stream(app_settings, payload, request)


async def api_chat_completions(app_settings: AppSettings, request: ChatCompletionRequest):
    """Creates a completion for the chat message"""
    error_check_ret = await check_model(app_settings, request, "generation", request.model)
    if error_check_ret is not None:
        return error_check_ret
    error_check_ret = check_requests(request)
    if error_check_ret is not None:
        return error_check_ret

    payload = get_gen_params(
        request.model,
        request.messages,
        temperature=request.temperature,
        top_p=request.top_p,
        max_tokens=request.max_tokens,
        echo=False,
        stream=request.stream,
        stop=request.stop,
    )
    if request.stream:
        return await chat_completions_stream(app_settings, payload, request)
    else:
        return await chat_completions_non_stream(app_settings, payload, request)

async def api_embeddings(app_settings: AppSettings, request: EmbeddingsRequest):
    """Creates embeddings for the text"""
    error_check_ret = await check_model(app_settings, request, "embedding", request.model)
    if error_check_ret is not None:
        return error_check_ret
    payload = {
        "model": request.model,
        "input": request.input,
    }

    response = await get_embedding(app_settings, payload)
    if response.type == "error":
        return create_error_response(ErrorCode.INTERNAL_ERROR, response.message)
    return EmbeddingsResponse(
        data=[EmbeddingsData(embedding=response.embedding, index=0)],
        model=request.model,
        usage=UsageInfo(
            prompt_tokens=response.usage.prompt_tokens,
            total_tokens=response.usage.total_tokens,
            completion_tokens=None,
        ),
    ).dict(exclude_none=True)


from typing import List, Optional, Union
import threading
import uuid

from fastapi import FastAPI, Request, BackgroundTasks
from fastapi.responses import StreamingResponse, JSONResponse
import requests
from tenacity import retry, stop_after_attempt
from langport.core.cluster_worker import ClusterWorker
from langport.model.executor.base import BaseModelExecutor
from langport.model.executor.embedding import EmbeddingExecutor

from langport.protocol.worker_protocol import (
    BaseWorkerResult,
    EmbeddingWorkerResult,
    EmbeddingsTask,
    UsageInfo,
)
import traceback

import torch

from langport.constants import (
    WORKER_API_TIMEOUT,
    WORKER_HEART_BEAT_INTERVAL,
    EMBEDDING_INFERENCE_INTERVAL,
    ErrorCode,
)
from langport.utils import server_error_msg, pretty_print_semaphore

class EmbeddingModelWorker(ClusterWorker):
    def __init__(
        self,
        node_addr: str,
        node_id: str,
        init_neighborhoods_addr: List[str],
        executor: EmbeddingExecutor,
        limit_model_concurrency: int,
        max_batch: int,
        stream_interval: int,
        logger
    ):
        super(EmbeddingModelWorker, self).__init__(
            node_addr=node_addr,
            node_id=node_id,
            init_neighborhoods_addr=init_neighborhoods_addr,
            limit_model_concurrency=limit_model_concurrency,
            max_batch=max_batch,
            stream_interval=stream_interval,
            logger=logger,
        )
        self.executor = executor
        workers = max(1, 2 * self.limit_model_concurrency // self.max_batch)
        self.add_timer(
            "embeddings_inference", 
            EMBEDDING_INFERENCE_INTERVAL, 
            executor.inference, 
            args=(self,), 
            kwargs=None, 
            workers=workers
        )
        
        self.on_start("set_features", self.set_features)
        self.on_start("set_model_name", self.set_model_name)

    async def set_features(self):
        await self.set_local_state("features", ["embedding"], ttl=360)
    
    async def set_model_name(self):
        await self.set_local_state("model_name", self.executor.model_name, ttl=360)

    async def get_embeddings(self, task: EmbeddingsTask):
        input_tokens = len(self.executor.tokenize(task.input))
        context_length = self.executor.context_length

        if input_tokens > context_length:
            return BaseWorkerResult(task_id=task.task_id,
                                   type="error",
                                   message=f"This model's maximum context length is {context_length} tokens. "
                                    f"However, you requested {input_tokens} tokens. "
                                    f"Please reduce the length of the messages or completion.",
                                    error_code=ErrorCode.CONTEXT_OVERFLOW
                                   )

        await self.add_task(task)
        result = None
        async for chunk in self.fetch_task_result(task.task_id):
            result = chunk
        return result



from langport.data.conversation import ConversationHistory, SeparatorStyle
from langport.model.model_adapter import BaseAdapter

from langport.data.conversation.conversation_settings import get_conv_settings

class BaizeAdapter(BaseAdapter):
    """The model adapter for project-baize/baize-lora-7B"""

    def match(self, model_path: str):
        return "baize" in model_path

    def get_default_conv_template(self, model_path: str) -> ConversationHistory:
        settings = get_conv_settings("baize")
        return ConversationHistory(
            system="The following is a conversation between a human and an AI assistant named Baize (named after a mythical creature in Chinese folklore). Baize is an open-source AI assistant developed by UCSD and Sun Yat-Sen University. The human and the AI assistant take turns chatting. Human statements start with [|Human|] and AI assistant statements start with [|AI|]. The AI assistant always provides responses in as much detail as possible, and in Markdown format. The AI assistant always declines to engage with topics, questions and instructions related to unethical, controversial, or sensitive issues. Complete the transcript in exactly that format.",
            messages=(
                ("[|Human|]", "Hello!"),
                ("[|AI|]", "Hi!"),
            ),
            offset=2,
            settings=settings,
        )




import json
from typing import Optional, Union
import httpx
import numpy as np
from pydantic import BaseSettings

from langport.core.dispatch import DispatchMethod
from langport.protocol.openai_api_protocol import ErrorResponse
from langport.protocol.worker_protocol import WorkerAddressRequest, WorkerAddressResponse

import json

from typing import Generator, Optional, Union, Dict, List, Any

from fastapi.responses import StreamingResponse, JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware, DispatchFunction
import httpx
import numpy as np
from pydantic import BaseSettings

from langport.constants import WORKER_API_TIMEOUT, ErrorCode
from langport.model.model_adapter import get_conversation_template
from fastapi.exceptions import RequestValidationError


class AppSettings(BaseSettings):
    # The address of the model controller.
    controller_address: str = "http://localhost:21001"

LANGPORT_HEADER = {"User-Agent": "Langport API Server"}

def create_error_response(code: int, message: str) -> JSONResponse:
    return JSONResponse(
        ErrorResponse(message=message, code=code).dict(), status_code=500
    )


async def _get_worker_address(
    app_settings: AppSettings, model_name: str, feature: str, client: httpx.AsyncClient, dispatch: Union[str, DispatchMethod]
) -> str:
    """
    Get worker address based on the requested model

    :param model_name: The worker's model name
    :param feature: The worker's feature
    :param client: The httpx client to use
    :return: Worker address from the controller
    :raises: :class:`ValueError`: No available worker for requested model
    """
    controller_address = app_settings.controller_address
    if isinstance(dispatch, str):
        dispatch = DispatchMethod.from_str(dispatch)
    if dispatch == DispatchMethod.LOTTERY:
        payload = WorkerAddressRequest(
            condition=f"{{model_name}}=='{model_name}' and '{feature}' in {{features}}", expression="1 / 0.01 + {speed}"
        )
    elif dispatch == DispatchMethod.SHORTEST_QUEUE:
        payload = WorkerAddressRequest(
            condition=f"{{model_name}}=='{model_name}' and '{feature}' in {{features}}", expression="{queue_length}/{speed}"
        )
    else:
        raise Exception("Error dispatch method.")
    ret = await client.post(
        controller_address + "/get_worker_address",
        json=payload.dict(),
    )
    response = WorkerAddressResponse.parse_obj(ret.json())
    address_list = response.address_list
    values = [json.loads(obj) for obj in response.values]

    # sort
    sorted_result = sorted(zip(address_list, values), key=lambda x: x[1])
    address_list = [x[0] for x in sorted_result]
    values = [x[1] for x in sorted_result]

    # No available worker
    if address_list == []:
        raise ValueError(f"No available worker for {model_name} and {feature}")
    if dispatch == DispatchMethod.LOTTERY:
        node_speeds = np.array(values, dtype=np.float32)
        norm = np.sum(node_speeds)
        if norm < 1e-4:
            return ""
        node_speeds = node_speeds / norm
        pt = np.random.choice(np.arange(len(address_list)), p=node_speeds)
        worker_addr = address_list[pt]
    elif dispatch == DispatchMethod.SHORTEST_QUEUE:
        worker_addr = address_list[0]
    else:
        raise Exception("Error dispatch method.")
    # logger.debug(f"model_name: {model_name}, feature: {feature}, worker_addr: {worker_addr}")
    return worker_addr


async def _list_models(app_settings: AppSettings, feature: Optional[str], client: httpx.AsyncClient) -> str:
    controller_address = app_settings.controller_address

    if feature is None:
        condition = "True"
    else:
        condition=f"'{feature}' in {{features}}"
    payload = WorkerAddressRequest(
        condition=condition, expression="{model_name}"
    )
    try:
        ret = await client.post(
            controller_address + "/get_worker_address",
            json=payload.dict(),
        )
        if ret.status_code != 200:
            return []
    except Exception as e:
        print("[Exception] list model: ", e)
        return []
    response = WorkerAddressResponse.parse_obj(ret.json())
    
    address_list = response.address_list
    models = [json.loads(obj) for obj in response.values]
    # No available worker
    if address_list == []:
        raise ValueError(f"No available worker for feature {feature}")

    return models

async def check_model(app_settings: AppSettings, request, feature: str, model_name: str) -> Optional[JSONResponse]:
    ret = None
    async with httpx.AsyncClient() as client:
        try:
            models = await _list_models(app_settings, feature, client)
        except Exception as e:
            ret = create_error_response(
                ErrorCode.INVALID_MODEL,
                str(e),
            )
            return ret
        if len(models) == 0 or model_name not in models:
            ret = create_error_response(
                ErrorCode.INVALID_MODEL,
                f"Only {'&&'.join(models)} allowed now, your model {model_name}",
            )
    return ret

def check_requests(request) -> Optional[JSONResponse]:
    # Check all params
    if request.max_tokens is not None and request.max_tokens <= 0:
        return create_error_response(
            ErrorCode.PARAM_OUT_OF_RANGE,
            f"{request.max_tokens} is less than the minimum of 1 - 'max_tokens'",
        )
    if request.n is not None and request.n <= 0:
        return create_error_response(
            ErrorCode.PARAM_OUT_OF_RANGE,
            f"{request.n} is less than the minimum of 1 - 'n'",
        )
    if request.temperature is not None and request.temperature < 0:
        return create_error_response(
            ErrorCode.PARAM_OUT_OF_RANGE,
            f"{request.temperature} is less than the minimum of 0 - 'temperature'",
        )
    if request.temperature is not None and request.temperature > 2:
        return create_error_response(
            ErrorCode.PARAM_OUT_OF_RANGE,
            f"{request.temperature} is greater than the maximum of 2 - 'temperature'",
        )
    if request.top_p is not None and request.top_p < 0:
        return create_error_response(
            ErrorCode.PARAM_OUT_OF_RANGE,
            f"{request.top_p} is less than the minimum of 0 - 'top_p'",
        )
    if request.top_p is not None and request.top_p > 1:
        return create_error_response(
            ErrorCode.PARAM_OUT_OF_RANGE,
            f"{request.top_p} is greater than the maximum of 1 - 'temperature'",
        )
    if request.stop is not None and (
        not isinstance(request.stop, str) and not isinstance(request.stop, list)
    ):
        return create_error_response(
            ErrorCode.PARAM_OUT_OF_RANGE,
            f"{request.stop} is not valid under any of the given schemas - 'stop'",
        )

    return None




from langport.data.conversation import (
    ConversationSettings,
    SeparatorStyle,
)


# RWKV-4-Raven default template
rwkv = ConversationSettings(
    name="rwkv",
    roles=("Bob", "Alice"),
    sep_style=SeparatorStyle.RWKV,
    sep="",
    stop_str="\n\n",
)


from typing import List, Optional
import warnings
from functools import cache

from langport.data.conversation import ConversationHistory
from langport.data.conversation.conversation_settings import get_conv_settings
from langport.model.model_adapter import BaseAdapter


class BiLLaAdapter(BaseAdapter):
    """The model adapter for BiLLa."""

    def match(self, model_path: str):
        return "billa" in model_path.lower()

    def get_default_conv_template(self, model_path: str) -> ConversationHistory:
        settings = get_conv_settings("bard")
        return ConversationHistory(
            system="",
            messages=(),
            offset=0,
            settings=settings,
        )



import argparse
import logging
import os

from langport.routers.gateway.common import AppSettings

logger = logging.getLogger(__name__)

import sys
from streamlit.web import cli as stcli

if __name__ in ["__main__", "langport.service.gateway.cluster_monitor"]:
    parser = argparse.ArgumentParser(
        description="Langport Cluster Monitor."
    )
    parser.add_argument("--host", type=str, default="localhost", help="host name")
    parser.add_argument("--port", type=int, default=8000, help="port number")
    parser.add_argument(
        "--controller-address", type=str, default="http://localhost:21001"
    )
    args = parser.parse_args()

    logger.debug(f"==== args ====\n{args}")

    app_settings = AppSettings(controller_address=args.controller_address)
    
    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, 'cluster_monitor_app.py')

    sys.argv = ["streamlit", "run", filename]
    sys.exit(stcli.main())



from langport.data.conversation import (
    ConversationSettings,
    SeparatorStyle,
)

# ChatGPT default template
chatgpt = ConversationSettings(
    name="chatgpt",
    roles=("user", "assistant"),
    sep_style=SeparatorStyle.ADD_COLON_SINGLE,
    sep="\n### ",
    stop_str="###",
)


# license: MIT (C) tardyp
import ast
import unittest


def safe_eval(expr, variables=None):
    """
    Safely evaluate a a string containing a Python
    expression.  The string or node provided may only consist of the following
    Python literal structures: strings, numbers, tuples, lists, dicts, booleans,
    and None. safe operators are allowed (and, or, ==, !=, not, +, -, ^, %, in, is)
    """
    _safe_names = {'None': None, 'True': True, 'False': False}
    _safe_nodes = [
        'Add', 'And', 'BinOp', 'BitAnd', 'BitOr', 'BitXor', 'BoolOp',
        'Compare', 'Dict', 'Eq', 'Expr', 'Expression', 'For',
        'Gt', 'GtE', 'Is', 'In', 'IsNot', 'LShift', 'List',
        'Load', 'Lt', 'LtE', 'Mod', 'Name', 'Not', 'NotEq', 'NotIn',
        'Num', 'Or', 'RShift', 'Set', 'Slice', 'Str', 'Sub',
        'Tuple', 'UAdd', 'USub', 'UnaryOp', 'boolop', 'cmpop',
        'expr', 'expr_context', 'operator', 'slice', 'unaryop', 'Constant', 'Div']
    node = ast.parse(expr, mode='eval')
    for subnode in ast.walk(node):
        subnode_name = type(subnode).__name__
        if isinstance(subnode, ast.Name):
            if subnode.id not in _safe_names and subnode.id not in variables:
                raise ValueError("Unsafe expression node {}. contains {}".format(expr, subnode.id))
        if subnode_name not in _safe_nodes:
            raise ValueError("Unsafe expression name {}. contains {}".format(expr, subnode_name))

    return eval(expr, variables)



class SafeEvalTests(unittest.TestCase):

    def test_basic(self):
        self.assertEqual(safe_eval("1", {}), 1)

    def test_local(self):
        self.assertEqual(safe_eval("a", {'a': 2}), 2)

    def test_local_bool(self):
        self.assertEqual(safe_eval("a==2", {'a': 2}), True)

    def test_lambda(self):
        self.assertRaises(ValueError, safe_eval, "lambda : None", {'a': 2})

    def test_bad_name(self):
        self.assertRaises(ValueError, safe_eval, "a == None2", {'a': 2})

    def test_attr(self):
        self.assertRaises(ValueError, safe_eval, "a.__dict__", {'a': 2})

    def test_eval(self):
        self.assertRaises(ValueError, safe_eval, "eval('os.exit()')", {})

    def test_exec(self):
        self.assertRaises(SyntaxError, safe_eval, "exec 'import os'", {})

    def test_multiply(self):
        self.assertRaises(ValueError, safe_eval, "'s' * 3", {})

    def test_power(self):
        self.assertRaises(ValueError, safe_eval, "3 ** 3", {})

    def test_comprehensions(self):
        self.assertRaises(ValueError, safe_eval, "[i for i in [1,2]]", {'i': 1})


from transformers import (
    AutoModel,
    AutoTokenizer,
)

from langport.model.model_adapter import BaseAdapter

class ChatGLMAdapter(BaseAdapter):
    """The model adapter for THUDM/chatglm-6b"""

    def match(self, model_path: str):
        return "chatglm" in model_path



from typing import Literal, Optional, List, Dict, Any, Union

import time

import shortuuid
from pydantic import BaseModel, Field

from langport.constants import ErrorCode

class NodeInfo(BaseModel):
    node_id: str
    node_addr: str
    check_heart_beat: bool
    refresh_time: int = Field(default_factory=lambda: int(time.time()))


class RegisterNodeRequest(BaseModel):
    node_id: str
    node_addr: str
    check_heart_beat: bool

class RegisterNodeResponse(BaseModel):
    node_id: str
    node_addr: str
    check_heart_beat: bool

class RemoveNodeRequest(BaseModel):
    node_id: str

class RemoveNodeResponse(BaseModel):
    node_id: str

class HeartbeatPing(BaseModel):
    node_id: str

class HeartbeatPong(BaseModel):
    exist: bool

class NodeListRequest(BaseModel):
    node_id: str

class NodeListResponse(BaseModel):
    nodes: List[NodeInfo]

class ListNodeStatesRequest(BaseModel):
    pass

class ListNodeStatesResponse(BaseModel):
    states: List[str]

class GetNodeStateRequest(BaseModel):
    state_name: str

class GetNodeStateResponse(BaseModel):
    state_value: str
    state_ttl: int = 60

class NodeInfoRequest(BaseModel):
    node_id: str

class NodeInfoResponse(BaseModel):
    node_info: NodeInfo

class WorkerAddressRequest(BaseModel):
    condition: str
    expression: str

class WorkerAddressResponse(BaseModel):
    id_list: List[str]
    address_list: List[str]
    values: List[str]


class BaseWorkerTask(BaseModel):
    task_id: str = Field(default_factory=lambda: f"task-{shortuuid.random()}")

class EmbeddingsTask(BaseWorkerTask):
    model: str
    input: str
    user: Optional[str] = None

class GenerationTask(BaseWorkerTask):
    prompt: str
    temperature: Optional[float] = 0.7
    repetition_penalty: Optional[float] = 0.0
    top_p: Optional[float] = 1.0
    top_k: Optional[int] = 1
    max_tokens: Optional[int] = None
    stop: Optional[Union[List[str], str]] = None
    echo: Optional[bool] = False
    stop_token_ids: Optional[List[int]] = None

class UsageInfo(BaseModel):
    prompt_tokens: int = 0
    total_tokens: int = 0
    completion_tokens: Optional[int] = 0

class BaseWorkerResult(BaseModel):
    task_id: str = Field(default_factory=lambda: f"task-{shortuuid.random()}")
    created: int = Field(default_factory=lambda: int(time.time()))
    type: Literal["data", "error", "finish", "done"]
    message: Optional[str] = None
    error_code: int = ErrorCode.OK

class EmbeddingWorkerResult(BaseWorkerResult):
    embedding: List[float]
    usage: UsageInfo = None

class GenerationWorkerResult(BaseWorkerResult):
    text: str
    logprobs: Optional[int] = None
    finish_reason: Optional[Literal["stop", "length"]] = None
    usage: UsageInfo = None



from typing import List, Optional
from ctransformers import AutoModelForCausalLM, LLM, AutoConfig, Config
from langport.model.executor.base import LocalModelExecutor
from langport.model.model_adapter import get_model_adapter

class GgmlTokenizer:
    def __init__(self, model:LLM) -> None:
        self.model = model
    
    def encode(self, text: str) -> List[int]:
        return self.model.tokenize(text)
    
    def decode(self, tokens: List[int]) -> str:
        return self.model.detokenize(tokens)
    
    def is_eos_token(self, token: int) -> bool:
        return self.model.is_eos_token(token)
        

class GgmlExecutor(LocalModelExecutor):
    def __init__(
        self,
        model_name: str,
        model_path: str,
        device: str,
        num_gpus: int,
        max_gpu_memory: Optional[str],
        lib: Optional[str] = None,
        gpu_layers: int = 0,
        model_type: str = 'llama',
        load_8bit: bool = False,
        cpu_offloading: bool = False,
    ) -> None:
        super(GgmlExecutor, self).__init__(
            model_name = model_name,
            model_path = model_path,
            device = device,
            num_gpus = num_gpus,
            max_gpu_memory = max_gpu_memory,
            load_8bit = load_8bit,
            cpu_offloading = cpu_offloading,
        )
        self.gpu_layers = gpu_layers
        # ctransformers has a bug
        self.lib = lib
        self.model_type = model_type
 

    def load_model(self, model_path: str, from_pretrained_kwargs: dict):
        adapter = get_model_adapter(model_path)
        config = Config()
        setattr(config, 'stream', True)
        setattr(config, 'gpu_layers', self.gpu_layers)
        auto_config = AutoConfig(config=config, model_type=self.model_type)
        model = AutoModelForCausalLM.from_pretrained(model_path,
                                                   config=auto_config,
                                                   lib=self.lib,
                                                   )
        tokenizer = GgmlTokenizer(model)

        return adapter, model, tokenizer



import argparse
import json
import logging

from typing import Optional

import fastapi
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware, DispatchFunction
from starlette.types import ASGIApp
import uvicorn

from langport.constants import ErrorCode
from fastapi.exceptions import RequestValidationError
from langport.protocol.openai_api_protocol import (
    ChatCompletionRequest,
    CompletionRequest,
    EmbeddingsRequest,
)
from langport.routers.gateway.common import AppSettings, create_error_response
from langport.routers.gateway.openai_compatible import api_chat_completions, api_completions, api_embeddings, api_models

logger = logging.getLogger(__name__)

app = fastapi.FastAPI(debug=False)

class BaseAuthorizationMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, sk:str, dispatch: Optional[DispatchFunction] = None) -> None:
        super().__init__(app, dispatch)
        self.sk = sk

    async def dispatch(self, request, call_next):
        authorization = request.headers.get("Authorization")
        if request.url.path not in ["/docs","/redoc","/openapi.json"] and (
            not authorization 
            or authorization.split(" ")[0].lower() != "bearer" 
            or authorization.split(" ")[1] != self.sk
        ):
            return JSONResponse(
                status_code=401,
                content={"msg":"Not authenticated"},
                headers={"WWW-Authenticate": "Bearer"},
            )
        return await call_next(request)


class RedirectModelMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp,  redirect_rules:list, dispatch: Optional[DispatchFunction] = None) -> None:
        super().__init__(app, dispatch)
        self.redirect_rules = redirect_rules
        self.receive_ = None

    async def dispatch(self, request, call_next):
        if "content-type" not in request.headers or request.headers["content-type"] != "application/json":
            return await call_next(request)
        
        try:
            await self.set_body(request)
            data = await request.json()
            if "model" in data:
                for rule in self.redirect_rules:
                    from_model_name, to_model_name = rule.split(":")
                    if data["model"] == from_model_name:
                        data["model"] = to_model_name
                        self.receive_['body'] = json.dumps(data).encode("utf-8")
                        break
        except Exception as e:
            logger.error(f"RedirectModelMiddleware: {e}")
        return await call_next(request)
    
    async def set_body(self, request):
        self.receive_ = await request._receive()
        async def receive():
            return self.receive_
        request._receive = receive


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request, exc):
    return create_error_response(ErrorCode.VALIDATION_TYPE_ERROR, str(exc))


@app.get("/v1/models")
async def models():
    return await api_models(app.app_settings)


@app.post("/v1/chat/completions")
async def chat_completions(request: ChatCompletionRequest):
    return await api_chat_completions(app.app_settings, request)

@app.post("/v1/completions")
async def completions(request: CompletionRequest):
    return await api_completions(app.app_settings, request)


@app.post("/v1/embeddings")
async def embeddings(request: EmbeddingsRequest):
    return await api_embeddings(app.app_settings, request)


if __name__ in ["__main__", "langport.service.gateway.openai_api"]:
    parser = argparse.ArgumentParser(
        description="Langport openai-compatible RESTful API server."
    )
    parser.add_argument("--host", type=str, default="localhost", help="host name")
    parser.add_argument("--port", type=int, default=8000, help="port number")
    parser.add_argument("--sk", type=str, default=None, help="security key")
    parser.add_argument(
        "--controller-address", type=str, default="http://localhost:21001"
    )
    parser.add_argument(
        "--allow-credentials", action="store_true", help="allow credentials"
    )
    parser.add_argument(
        "--allowed-origins", type=json.loads, default=["*"], help="allowed origins"
    )
    parser.add_argument(
        "--allowed-methods", type=json.loads, default=["*"], help="allowed methods"
    )
    parser.add_argument(
        "--allowed-headers", type=json.loads, default=["*"], help="allowed headers"
    )
    parser.add_argument("--redirect", action="append", required=False, help="redirect model_name to another model_name, e.g. --redirect model_name1:model_name2")
    args = parser.parse_args()

    app.add_middleware(
        CORSMiddleware,
        allow_origins=args.allowed_origins,
        allow_credentials=args.allow_credentials,
        allow_methods=args.allowed_methods,
        allow_headers=args.allowed_headers,
    )
    if args.redirect is not None:
        app.add_middleware(
            RedirectModelMiddleware,
            redirect_rules=args.redirect,
        )
    if args.sk is not None:
        app.add_middleware(
            BaseAuthorizationMiddleware,
            sk=args.sk,
        )
    app.app_settings = AppSettings(controller_address=args.controller_address)

    logger.debug(f"==== args ====\n{args}")

    # don't delete this line, otherwise the middleware won't work with reload==True
    if __name__ == "__main__":
        uvicorn.run(
            "langport.service.gateway.openai_api:app",
            host=args.host,
            port=args.port,
            log_level="info",
            reload=True,
        )



import argparse
import json
import logging

from typing import Optional, Union

import fastapi
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware, DispatchFunction
from starlette.types import ASGIApp
import uvicorn

from langport.constants import ErrorCode
from fastapi.exceptions import RequestValidationError
from langport.protocol.tabby_api_protocol import (
    Choice,
    CompletionRequest,
    CompletionResponse,
    ChoiceEvent,
    CompletionEvent,
    EventTypeMapping,
    HTTPValidationError,
    LanguagePresets,
)
from langport.protocol.openai_api_protocol import CompletionRequest as OpenAICompletionRequest

from langport.routers.gateway.common import AppSettings, check_model, create_error_response
from langport.routers.gateway.openai_compatible import completions_non_stream, get_gen_params

logger = logging.getLogger(__name__)

app = fastapi.FastAPI(debug=False)

class BaseAuthorizationMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, sk:str, dispatch: Optional[DispatchFunction] = None) -> None:
        super().__init__(app, dispatch)
        self.sk = sk

    async def dispatch(self, request, call_next):
        authorization = request.headers.get("Authorization")
        if request.url.path not in ["/docs","/redoc","/openapi.json"] and (
            not authorization 
            or authorization.split(" ")[0].lower() != "bearer" 
            or authorization.split(" ")[1] != self.sk
        ):
            return JSONResponse(
                status_code=401,
                content={"msg":"Not authenticated"},
                headers={"WWW-Authenticate": "Bearer"},
            )
        return await call_next(request)


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request, exc):
    return create_error_response(ErrorCode.VALIDATION_TYPE_ERROR, str(exc))

def trim_with_stop_words(output: str, stopwords: list) -> str:
    for w in sorted(stopwords, key=len, reverse=True):
        index = output.find(w)
        if index != -1:
            output = output[:index]
    return output

@app.post("/v1/events")
async def events(e: Union[ChoiceEvent, CompletionEvent]):
    if isinstance(e, EventTypeMapping[e.type]):
        # events_lib.log_event(e)
        return JSONResponse(content="ok")
    else:
        print(type(e))
        return JSONResponse(content="invalid event", status_code=422)


@app.post("/v1/completions")
async def completions(request: CompletionRequest):
    error_check_ret = await check_model(app.app_settings, request, "generation", app.model_name)
    if error_check_ret is not None:
        return error_check_ret
    
    preset = LanguagePresets.get(request.language, None)
    if preset is None:
        return CompletionResponse()

    # print(request.prompt)
    payload = get_gen_params(
        app.model_name,
        request.prompt,
        temperature=0.7,
        top_p=1.0,
        max_tokens=preset.max_length,
        echo=False,
        stream=False,
        stop=preset.stop_words,
    )
    N = 1
    response = await completions_non_stream(
        app.app_settings,
        payload,
        OpenAICompletionRequest(
            model=app.model_name,
            prompt=request.prompt,
            n=N,
        )
    )
    if isinstance(response, JSONResponse):
        print(response.body)
        return CompletionResponse()

    # print(response.choices[0].text.replace("\n", "\\n"))
    return CompletionResponse(choices=[
        Choice(
            index=i,
            text=response.choices[i].text
        )
        for i in range(len(response.choices))
    ])

if __name__ in ["__main__", "langport.service.gateway.tabby_api"]:
    parser = argparse.ArgumentParser(
        description="Langport tabby-compatible RESTful API server."
    )
    parser.add_argument("--host", type=str, default="localhost", help="host name")
    parser.add_argument("--port", type=int, default=8000, help="port number")
    parser.add_argument("--model-name", type=str, default="J-350M", help="default tabby model name")
    parser.add_argument("--sk", type=str, default=None, help="security key")
    parser.add_argument(
        "--controller-address", type=str, default="http://localhost:21001"
    )
    parser.add_argument(
        "--allow-credentials", action="store_true", help="allow credentials"
    )
    parser.add_argument(
        "--allowed-origins", type=json.loads, default=["*"], help="allowed origins"
    )
    parser.add_argument(
        "--allowed-methods", type=json.loads, default=["*"], help="allowed methods"
    )
    parser.add_argument(
        "--allowed-headers", type=json.loads, default=["*"], help="allowed headers"
    )
    args = parser.parse_args()

    app.add_middleware(
        CORSMiddleware,
        allow_origins=args.allowed_origins,
        allow_credentials=args.allow_credentials,
        allow_methods=args.allowed_methods,
        allow_headers=args.allowed_headers,
    )
    if args.sk is not None:
        app.add_middleware(
            BaseAuthorizationMiddleware,
            sk=args.sk,
        )
    app.model_name = args.model_name
    app.app_settings = AppSettings(controller_address=args.controller_address)

    logger.debug(f"==== args ====\n{args}")

    # don't delete this line, otherwise the middleware won't work with reload==True
    if __name__ == "__main__":
        uvicorn.run(
            "langport.service.gateway.tabby_api:app",
            host=args.host,
            port=args.port,
            log_level="info",
            reload=True,
        )



import json
import pandas as pd
import streamlit as st
import asyncio
import httpx
from langport.protocol.worker_protocol import WorkerAddressRequest, WorkerAddressResponse

from langport.routers.gateway.common import AppSettings, _list_models

async def list_workers(app_settings: AppSettings):
    data_dict = {
        "address": [],
        "features": [],
        "model": [],
        "queue_length": [],
    }
    async with httpx.AsyncClient() as client:
        payload = WorkerAddressRequest(
            condition="True", expression="{model_name}, {features}, {queue_length}"
        )
        ret = await client.post(
            app_settings.controller_address + "/get_worker_address",
            json=payload.dict(),
        )
        if ret.status_code != 200:
            return []
        response = WorkerAddressResponse.parse_obj(ret.json())
        address_list = response.address_list
        data = [json.loads(obj) for obj in response.values]

    for i in range(len(address_list)):
        data_dict["address"].append(address_list[i])
        data_dict["features"].append(str(data[i][1]))
        data_dict["model"].append(data[i][0])
        data_dict["queue_length"].append(data[i][2])
    
    df = pd.DataFrame(data_dict)
    return df

async def main():
    app_settings = AppSettings(controller_address="http://localhost:21001")
    st.title('LangPort Cluster Monitor')

    df = await list_workers(app_settings)
    st.table(df)
if __name__ == "__main__":
    asyncio.run(main())


from transformers import (
    AutoModelForSeq2SeqLM,
    T5Tokenizer,
)

from langport.model.model_adapter import BaseAdapter


class T5Adapter(BaseAdapter):
    """The model adapter for lmsys/fastchat-t5-3b-v1.0"""

    def match(self, model_path: str):
        return "t5" in model_path




from fastapi import FastAPI, Request, BackgroundTasks
from fastapi.responses import StreamingResponse, JSONResponse

from langport.protocol.worker_protocol import EmbeddingsTask, GenerationTask
from .core_node import app, create_background_tasks


@app.post("/chat_stream")
async def api_chat_stream(request: Request):
    params = await request.json()
    await app.node.acquire_model_semaphore()
    generator = app.node.generation_bytes_stream(GenerationTask(
        prompt=params["prompt"],
        temperature=params.get("temperature", 1.0),
        repetition_penalty=params.get("repetition_penalty", 0.0),
        top_p=params.get("top_p", 1.0),
        top_k=params.get("top_k", 1),
        max_tokens=params.get("max_tokens", 512),
        stop=params.get("stop", None),
        echo=params.get("echo", False),
        stop_token_ids=params.get("stop_token_ids", None),
    ))
    background_tasks = create_background_tasks(app.node)
    return StreamingResponse(generator, background=background_tasks)

@app.post("/chat")
async def api_chat(request: Request):
    params = await request.json()
    await app.node.acquire_model_semaphore()
    generator = await app.node.generation_stream(GenerationTask(
        prompt=params["prompt"],
        temperature=params.get("temperature", 1.0),
        repetition_penalty=params.get("repetition_penalty", 0.0),
        top_p=params.get("top_p", 1.0),
        top_k=params.get("top_k", 1),
        max_tokens=params.get("max_tokens", 512),
        stop=params.get("stop", None),
        echo=params.get("echo", False),
        stop_token_ids=params.get("stop_token_ids", None),
    ))
    completion = None
    for chunk in generator:
        completion = chunk
    background_tasks = create_background_tasks(app.node)
    return JSONResponse(content=completion.dict(), background=background_tasks)



@app.post("/completion_stream")
async def api_completion_stream(request: Request):
    params = await request.json()
    await app.node.acquire_model_semaphore()
    generator = app.node.generation_bytes_stream(GenerationTask(
        prompt=params["prompt"],
        temperature=params.get("temperature", 1.0),
        repetition_penalty=params.get("repetition_penalty", 0.0),
        top_p=params.get("top_p", 1.0),
        top_k=params.get("top_k", 1),
        max_tokens=params.get("max_tokens", 512),
        stop=params.get("stop", None),
        echo=params.get("echo", False),
        stop_token_ids=params.get("stop_token_ids", None),
    ))
    background_tasks = create_background_tasks(app.node)
    return StreamingResponse(generator, background=background_tasks)


@app.post("/completion")
async def api_completion(request: Request):
    params = await request.json()
    await app.node.acquire_model_semaphore()
    generator = await app.node.generation_stream(GenerationTask(
        prompt=params["prompt"],
        temperature=params.get("temperature", 1.0),
        repetition_penalty=params.get("repetition_penalty", 0.0),
        top_p=params.get("top_p", 1.0),
        top_k=params.get("top_k", 1),
        max_tokens=params.get("max_tokens", 512),
        stop=params.get("stop", None),
        echo=params.get("echo", False),
        stop_token_ids=params.get("stop_token_ids", None),
    ))
    completion = None
    for chunk in generator:
        completion = chunk
    background_tasks = create_background_tasks(app.node)
    return JSONResponse(content=completion.dict(), background=background_tasks)



from langport.data.conversation import (
    ConversationSettings,
    SeparatorStyle,
)


# Claude default template
claude = ConversationSettings(
    name="claude",
    roles=("Human", "Assistant"),
    sep_style=SeparatorStyle.ADD_COLON_SINGLE,
    sep="\n\n",
)



import argparse
import json
import logging

from typing import Optional
import fastapi
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware, DispatchFunction
from starlette.types import ASGIApp
import uvicorn

from langport.constants import ErrorCode
from fastapi.exceptions import RequestValidationError
from langport.protocol.huggingface_api_protocol import Details, FinishReason, Request, Response
from langport.protocol.openai_api_protocol import CompletionRequest as OpenAICompletionRequest

from langport.routers.gateway.common import AppSettings, check_model, create_error_response
from langport.routers.gateway.openai_compatible import completions_non_stream, get_gen_params

logger = logging.getLogger(__name__)

app = fastapi.FastAPI(debug=False)

class BaseAuthorizationMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, sk:str, dispatch: Optional[DispatchFunction] = None) -> None:
        super().__init__(app, dispatch)
        self.sk = sk

    async def dispatch(self, request, call_next):
        authorization = request.headers.get("Authorization")
        if request.url.path not in ["/docs","/redoc","/openapi.json"] and (
            not authorization 
            or authorization.split(" ")[0].lower() != "bearer" 
            or authorization.split(" ")[1] != self.sk
        ):
            return JSONResponse(
                status_code=401,
                content={"msg":"Not authenticated"},
                headers={"WWW-Authenticate": "Bearer"},
            )
        return await call_next(request)


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request, exc):
    return create_error_response(ErrorCode.VALIDATION_TYPE_ERROR, str(exc))


@app.post("/")
async def root(request: Request):
    pass


@app.post("/generate")
async def generate(request: Request):
    error_check_ret = await check_model(app.app_settings, request, "generation", app.model_name)
    if error_check_ret is not None:
        return error_check_ret
    
    temperature = 0.7
    top_p = 1.0
    top_k = 1.0
    max_tokens = 2048
    stop = []
    if request.parameters is not None:
        temperature = request.parameters.temperature
        top_p = request.parameters.top_p
        top_k = request.parameters.top_k
        max_tokens = request.parameters.max_new_tokens
        stop = request.parameters.stop
    payload = get_gen_params(
        app.model_name,
        request.inputs,
        temperature=temperature,
        top_p=top_p,
        max_tokens=max_tokens,
        echo=False,
        stream=False,
        stop=stop,
    )
    N = 1
    response = await completions_non_stream(
        app.app_settings,
        payload,
        OpenAICompletionRequest(
            model=app.model_name,
            prompt=request.inputs,
            n=N,
        )
    )
    if isinstance(response, JSONResponse):
        print(response.body)
        return Response(generated_text="", details=Details(
            finish_reason=FinishReason.EndOfSequenceToken,
            generated_tokens=0,
        ))

    # print(response.choices[0].text.replace("\n", "\\n"))
    return Response(generated_text=response.choices[0].text,
            details=Details(
                finish_reason=FinishReason.EndOfSequenceToken,
                generated_tokens=response.usage.completion_tokens,
            )
        )
    


@app.post("/generate_stream")
async def generate_stream(request: Request):
    error_check_ret = await check_model(app.app_settings, request, "generation", app.model_name)
    if error_check_ret is not None:
        return error_check_ret


if __name__ in ["__main__", "langport.service.gateway.huggingface_api"]:
    parser = argparse.ArgumentParser(
        description="Langport huggingface-compatible RESTful API server."
    )
    parser.add_argument("--host", type=str, default="localhost", help="host name")
    parser.add_argument("--port", type=int, default=8000, help="port number")
    parser.add_argument("--sk", type=str, default=None, help="security key")
    parser.add_argument("--model-name", type=str, default="J-350M", help="default tabby model name")
    parser.add_argument(
        "--controller-address", type=str, default="http://localhost:21001"
    )
    parser.add_argument(
        "--allow-credentials", action="store_true", help="allow credentials"
    )
    parser.add_argument(
        "--allowed-origins", type=json.loads, default=["*"], help="allowed origins"
    )
    parser.add_argument(
        "--allowed-methods", type=json.loads, default=["*"], help="allowed methods"
    )
    parser.add_argument(
        "--allowed-headers", type=json.loads, default=["*"], help="allowed headers"
    )
    args = parser.parse_args()

    app.add_middleware(
        CORSMiddleware,
        allow_origins=args.allowed_origins,
        allow_credentials=args.allow_credentials,
        allow_methods=args.allowed_methods,
        allow_headers=args.allowed_headers,
    )
    if args.sk is not None:
        app.add_middleware(
            BaseAuthorizationMiddleware,
            sk=args.sk,
        )
    app.model_name = args.model_name
    app.app_settings = AppSettings(controller_address=args.controller_address)

    logger.debug(f"==== args ====\n{args}")

    # don't delete this line, otherwise the middleware won't work with reload==True
    if __name__ == "__main__":
        uvicorn.run(
            "langport.service.gateway.huggingface_api:app",
            host=args.host,
            port=args.port,
            log_level="info",
            reload=True,
        )



import argparse
import json
import logging

from typing import Optional

import fastapi
import uvicorn
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware, DispatchFunction
from starlette.types import ASGIApp

from langport.constants import ErrorCode
from fastapi.exceptions import RequestValidationError
from langport.protocol.openai_api_protocol import (
    CompletionRequest,
)
from langport.routers.gateway.common import AppSettings, create_error_response
from langport.routers.gateway.openai_compatible import api_completions, api_models


logger = logging.getLogger(__name__)

app = fastapi.FastAPI(debug=False)

class BaseAuthorizationMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, sk:str, dispatch: Optional[DispatchFunction] = None) -> None:
        super().__init__(app, dispatch)
        self.sk = sk

    async def dispatch(self, request, call_next):
        authorization = request.headers.get("Authorization")
        if request.url.path not in ["/docs","/redoc","/openapi.json"] and (
            not authorization 
            or authorization.split(" ")[0].lower() != "bearer" 
            or authorization.split(" ")[1] != self.sk
        ):
            return JSONResponse(
                status_code=401,
                content={"msg":"Not authenticated"},
                headers={"WWW-Authenticate": "Bearer"},
            )
        return await call_next(request)


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request, exc):
    return create_error_response(ErrorCode.VALIDATION_TYPE_ERROR, str(exc))


@app.get("/v1/models")
async def models():
    return await api_models(app.app_settings)

@app.post("/v1/engines/codegen/completions")
@app.post("/v1/engines/copilot-codex/completions")
@app.post("/v1/completions")
async def completions(request: CompletionRequest):
    return await api_completions(app.app_settings, request)

if __name__ in ["__main__", "langport.service.gateway.fauxpilot_api"]:
    parser = argparse.ArgumentParser(
        description="Langport fauxpilot-compatible RESTful API server."
    )
    parser.add_argument("--host", type=str, default="localhost", help="host name")
    parser.add_argument("--port", type=int, default=8000, help="port number")
    parser.add_argument("--sk", type=str, default=None, help="security key")
    parser.add_argument(
        "--controller-address", type=str, default="http://localhost:21001"
    )
    parser.add_argument(
        "--allow-credentials", action="store_true", help="allow credentials"
    )
    parser.add_argument(
        "--allowed-origins", type=json.loads, default=["*"], help="allowed origins"
    )
    parser.add_argument(
        "--allowed-methods", type=json.loads, default=["*"], help="allowed methods"
    )
    parser.add_argument(
        "--allowed-headers", type=json.loads, default=["*"], help="allowed headers"
    )
    args = parser.parse_args()

    app.add_middleware(
        CORSMiddleware,
        allow_origins=args.allowed_origins,
        allow_credentials=args.allow_credentials,
        allow_methods=args.allowed_methods,
        allow_headers=args.allowed_headers,
    )
    if args.sk is not None:
        app.add_middleware(
            BaseAuthorizationMiddleware,
            sk=args.sk,
        )
    app.app_settings = AppSettings(controller_address=args.controller_address)

    logger.debug(f"==== args ====\n{args}")

    # don't delete this line, otherwise the middleware won't work with reload==True
    if __name__ == "__main__":
        uvicorn.run(
            "langport.service.gateway.fauxpilot_api:app",
            host=args.host,
            port=args.port,
            log_level="info",
            reload=True,
        )



from langport.data.conversation import ConversationHistory, SeparatorStyle
from langport.data.conversation.conversation_settings import get_conv_settings
from langport.model.model_adapter import BaseAdapter


class KoalaAdapter(BaseAdapter):
    """The model adapter for koala"""

    def match(self, model_path: str):
        return "koala" in model_path

    def get_default_conv_template(self, model_path: str) -> ConversationHistory:
        settings = get_conv_settings("dolly_v2")
        return ConversationHistory(
            system="BEGINNING OF CONVERSATION:",
            messages=(),
            offset=0,
            settings=settings,
        )



import argparse
import os
import random
import uuid
import uvicorn

from langport.workers.embedding_worker import EmbeddingModelWorker
from langport.model.executor.embedding.huggingface import HuggingfaceEmbeddingExecutor
from langport.model.model_args import add_model_args
from langport.utils import build_logger
from langport.routers.server.embedding_node import app

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", type=str, default="localhost")
    parser.add_argument("--port", type=int, default=None)
    parser.add_argument("--worker-address", type=str, default=None)
    parser.add_argument("--neighbors", type=str, nargs="*", default=[])
    
    add_model_args(parser)
    parser.add_argument("--model-name", type=str, help="Optional display name")
    parser.add_argument("--limit-model-concurrency", type=int, default=8)
    parser.add_argument("--batch", type=int, default=4)
    parser.add_argument("--stream-interval", type=int, default=2)
    parser.add_argument(
        "--dispatch-method",
        type=str,
        choices=["lottery", "shortest_queue"],
        default="shortest_queue",
    )
    args = parser.parse_args()

    node_id = str(uuid.uuid4())
    logger = build_logger("embedding_worker", f"embedding_worker_{node_id}.log")
    logger.info(f"args: {args}")

    if args.gpus:
        if len(args.gpus.split(",")) < args.num_gpus:
            raise ValueError(
                f"Larger --num-gpus ({args.num_gpus}) than --gpus {args.gpus}!"
            )
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpus

    if args.port is None:
        args.port = random.randint(21001, 29001)

    if args.worker_address is None:
        args.worker_address = f"http://{args.host}:{args.port}"
    
    if args.model_name is None:
        args.model_name = os.path.basename(os.path.normpath(args.model_path))

    executor = HuggingfaceEmbeddingExecutor(
        model_name=args.model_name,
        model_path=args.model_path,
        device=args.device,
        num_gpus=args.num_gpus,
        max_gpu_memory=args.max_gpu_memory,
        load_8bit=args.load_8bit,
        cpu_offloading=args.cpu_offloading,
        deepspeed=args.deepspeed,
        trust_remote_code=args.trust_remote_code
    )

    app.node = EmbeddingModelWorker(
        node_addr=args.worker_address,
        node_id=node_id,
        init_neighborhoods_addr=args.neighbors,
        executor=executor,
        limit_model_concurrency=args.limit_model_concurrency,
        max_batch=args.batch,
        stream_interval=args.stream_interval,
        logger=logger,
    )
    uvicorn.run(app, host=args.host, port=args.port, log_level="info")



from langport.data.conversation import (
    ConversationSettings,
    SeparatorStyle,
)


h2ogpt = ConversationSettings(
    name="h2ogpt",
    roles=("<|prompt|>", "<|answer|>"),
    sep_style=SeparatorStyle.NO_COLON_SINGLE,
    sep="</s>",
)


from langport.data.conversation import (
    ConversationSettings,
    SeparatorStyle,
)


# Dolly V2 default template
dolly_v2 = ConversationSettings(
    name="dolly_v2",
    roles=("### Instruction", "### Response"),
    sep_style=SeparatorStyle.DOLLY,
    sep="\n\n",
    sep2="### End",
)


from langport.model.model_adapter import BaseAdapter

class BertAdapter(BaseAdapter):
    """The model adapter for bert text2vec models."""

    def match(self, model_path: str):
        return "bert" in model_path





from enum import Enum, auto

class DispatchMethod(Enum):
    LOTTERY = auto()
    SHORTEST_QUEUE = auto()

    @classmethod
    def from_str(cls, name):
        if name == "lottery":
            return cls.LOTTERY
        elif name == "shortest_queue":
            return cls.SHORTEST_QUEUE
        else:
            raise ValueError(f"Invalid dispatch method")



from typing import List, Optional
from langport.model.executor.ggml import GgmlExecutor, GgmlTokenizer
from ctransformers import LLM
from langport.protocol.worker_protocol import BaseWorkerResult, GenerationTask, GenerationWorkerResult, UsageInfo
from langport.workers.generation_worker import GenerationModelWorker


def stream_generation(
    model: LLM,
    tokenizer: GgmlTokenizer,
    stream_interval: int,
    tasks: List[GenerationTask],
):
    batch_size = len(tasks)
    if batch_size == 0:
        return

    # todo: add stop words support
    for i, task in enumerate(tasks):
        output = ""

        if task.echo:
            output = task.prompt
        else:
            tokens = tokenizer.encode(task.prompt)
            prompt_length = len(tokens)
            output_ids = []

            # Compatible with some models
            top_k = 40 if task.top_k <= 1 else task.top_k
            repetition_penalty = 1.17647 if task.repetition_penalty == 0.0 else task.repetition_penalty

            for j, token in enumerate(model.generate(tokens, top_k=top_k, top_p=task.top_p,
                                                  temperature=task.temperature, repetition_penalty=repetition_penalty)):
                output_ids.append(token)
                if tokenizer.is_eos_token(token) or prompt_length + j == task.max_tokens - 1:
                    output = tokenizer.decode(output_ids)
                    if tokenizer.is_eos_token(token):
                        finish_reason = "stop"
                    else:
                        finish_reason = "length"
                    yield GenerationWorkerResult(
                        task_id=task.task_id,
                        type="finish",
                        text=output,
                        usage=UsageInfo(
                            prompt_tokens=prompt_length,
                            total_tokens=prompt_length + j,
                            completion_tokens=j,
                        ),
                        finish_reason=finish_reason,
                    )
                    break

                if j%stream_interval!=0:
                    continue
                output = tokenizer.decode(output_ids)

                # yield result
                yield GenerationWorkerResult(
                    task_id=task.task_id,
                    type="data",
                    text=output,
                    usage=UsageInfo(
                        prompt_tokens=prompt_length,
                        total_tokens=prompt_length + j,
                        completion_tokens=j,
                    ),
                    finish_reason=None,
                )


class GgmlGenerationExecutor(GgmlExecutor):
    def __init__(
        self,
        model_name: str,
        model_path: str,
        context_length: int,
        gpu_layers: int,
        model_type: str = "llama",
        lib: Optional[str] = None,
    ) -> None:
        n_gpu = 1 if gpu_layers > 0 else 0
        super(GgmlGenerationExecutor, self).__init__(
            model_name=model_name,
            model_path=model_path,
            device="cpu",
            num_gpus=n_gpu,
            max_gpu_memory=None,
            gpu_layers=gpu_layers,
            lib=lib,
            model_type=model_type,
        )
        self.n_ctx = context_length
        self.adapter, self.model, self.tokenizer = self.load_model(model_path, from_pretrained_kwargs={})

    @property
    def context_length(self) -> int:
        return self.n_ctx

    def tokenize(self, text: str) -> List[int]:
        return self.tokenizer.encode(text)
    
    def inference(self, worker: "GenerationModelWorker"):
        if not worker.online:
            return

        tasks = worker.fetch_tasks()
        batch_size = len(tasks)
        if batch_size == 0:
            return

        # batch inference
        for chunk in stream_generation(
            self.model,
            self.tokenizer,
            worker.stream_interval,
            tasks,
        ):
            worker.push_task_result(chunk.task_id, chunk)

        for task in tasks:
            worker.push_task_result(
                task.task_id, BaseWorkerResult(task_id=task.task_id, type="done")
            )
    


import os
from langport.data.conversation import ConversationHistory
from langport.data.conversation.conversation_settings import get_conv_settings
from langport.model.model_adapter import BaseAdapter

class OpenBuddyAdapter(BaseAdapter):
    """The model adapter for OpenBuddy"""

    def match(self, model_path: str):
        if model_path.endswith('/'):
            model_path = model_path[:-1]
        dir_name = os.path.basename(model_path)
        parent_name = os.path.dirname(model_path)
        return dir_name.startswith('openbuddy') or parent_name.startswith('openbuddy')

    def get_default_conv_template(self, model_path: str) -> ConversationHistory:
        settings = get_conv_settings("openbuddy")
        return ConversationHistory(
            system="""Consider a conversation between User (a human) and Assistant (named Buddy).
Buddy is an INTP-T, a friendly, intelligent and multilingual AI assistant, by OpenBuddy team. GitHub: https://github.com/OpenBuddy/OpenBuddy
Buddy cannot access the Internet.
Buddy can fluently speak the user's language (e.g. English, Chinese).
Buddy can generate poems, stories, code, essays, songs, parodies, and more.
Buddy possesses vast knowledge about the world, history, and culture.
Buddy's responses are always safe, creative, high-quality, human-like, and interesting.
Buddy strictly refuses to discuss political, NSFW, or other unsafe topics.

User: Hi.
Assistant: Hi, I'm Buddy, your AI assistant. How can I help you today?""",
            messages=[],
            offset=0,
            settings=settings,
        )



from langport.data.conversation import (
    ConversationSettings,
    SeparatorStyle,
)


# StableLM Alpha default template
stablelm = ConversationSettings(
    name="stablelm",
    roles=("<|USER|>", "<|ASSISTANT|>"),
    sep_style=SeparatorStyle.NO_COLON_SINGLE,
    sep="",
    stop_token_ids=[50278, 50279, 50277, 1, 0],
)



from enum import IntEnum


HEART_BEAT_EXPIRATION = 90
WORKER_HEART_BEAT_INTERVAL = 30
EMBEDDING_INFERENCE_INTERVAL = 0.1
GENERATION_INFERENCE_INTERVAL = 0.3
WORKER_API_TIMEOUT = 20

LOGDIR = "./logs"

class ErrorCode(IntEnum):
    OK = 200
    VALIDATION_TYPE_ERROR = 40001

    INVALID_AUTH_KEY = 40101
    INCORRECT_AUTH_KEY = 40102
    NO_PERMISSION = 40103

    INVALID_MODEL = 40301
    PARAM_OUT_OF_RANGE = 40302
    CONTEXT_OVERFLOW = 40303

    RATE_LIMIT = 42901
    QUOTA_EXCEEDED = 42902
    ENGINE_OVERLOADED = 42903

    INTERNAL_ERROR = 50001
    CUDA_OUT_OF_MEMORY = 50002



from transformers import AutoModelForCausalLM, AutoTokenizer
from langport.model.model_adapter import BaseAdapter

class CodeGenAdapter(BaseAdapter):
    """The model adapter for CodeGen models."""

    def match(self, model_path: str):
        return "codegen" in model_path


class CodeGen2Adapter(BaseAdapter):
    """The model adapter for CodeGen2 models."""

    def match(self, model_path: str):
        return "codegen2" in model_path



"""Model adapter registration."""

import importlib
import pkgutil
import sys
import os
from typing import List
import warnings
from functools import cache

from langport.data.conversation import ConversationHistory, SeparatorStyle
from langport.data.conversation.conversation_settings import get_conv_settings


class BaseAdapter:
    """The base and the default model adapter."""

    def match(self, model_path: str) -> bool:
        return True

    def get_default_conv_template(self, model_path: str) -> ConversationHistory:
        settings = get_conv_settings("zero_shot")
        return ConversationHistory(
            system="A chat between a curious human and an artificial intelligence assistant. "
            "The assistant gives helpful, detailed, and polite answers to the human's questions.",
            messages=[],
            offset=2,
            settings=settings,
        )


# A global registry for all model adapters
model_adapters: List[BaseAdapter] = []


def register_model_adapter(cls):
    """Register a model adapter."""
    model_adapters.append(cls())


@cache
def get_model_adapter(model_path: str) -> BaseAdapter:
    """Get a model adapter for a model_path."""
    for adapter in model_adapters:
        if adapter.match(model_path):
            print(f"Using model adapter {adapter.__class__.__name__}")
            return adapter
    raise ValueError(f"No valid model adapter for {model_path}")


def raise_warning_for_incompatible_cpu_offloading_configuration(
    device: str, load_8bit: bool, cpu_offloading: bool
):
    if cpu_offloading:
        if not load_8bit:
            warnings.warn(
                "The cpu-offloading feature can only be used while also using 8-bit-quantization.\n"
                "Use '--load-8bit' to enable 8-bit-quantization\n"
                "Continuing without cpu-offloading enabled\n"
            )
            return False
        if not "linux" in sys.platform:
            warnings.warn(
                "CPU-offloading is only supported on linux-systems due to the limited compatability with the bitsandbytes-package\n"
                "Continuing without cpu-offloading enabled\n"
            )
            return False
        if device != "cuda":
            warnings.warn(
                "CPU-offloading is only enabled when using CUDA-devices\n"
                "Continuing without cpu-offloading enabled\n"
            )
            return False
    return cpu_offloading


def get_conversation_template(model_path: str) -> ConversationHistory:
    adapter = get_model_adapter(model_path)
    return adapter.get_default_conv_template(model_path)

adapters_path = os.path.join(os.path.dirname(__file__), "adapters")
for module_loader, name, ispkg in pkgutil.iter_modules([adapters_path]):
    # print(module_loader, name, ispkg)
    importlib.import_module(".adapters." + name, __package__)

# Register all adapters.
for cls in BaseAdapter.__subclasses__():
    register_model_adapter(cls)

# After all adapters, try the default base adapter.
register_model_adapter(BaseAdapter)



from fastapi import FastAPI, Request, BackgroundTasks
from fastapi.responses import StreamingResponse, JSONResponse

from langport.protocol.worker_protocol import EmbeddingsTask
from .core_node import app, create_background_tasks


@app.post("/embeddings")
async def api_embeddings(request: EmbeddingsTask):
    await app.node.acquire_model_semaphore()
    embedding = await app.node.get_embeddings(request)
    background_tasks = create_background_tasks(app.node)
    return JSONResponse(content=embedding.dict(), background=background_tasks)





import importlib
import os
import pkgutil
from typing import Dict

from langport.data.conversation import ConversationHistory, ConversationSettings


# A global registry for all conversation settings
conv_settings: Dict[str, ConversationSettings] = {}


def register_conv_settings(template: ConversationSettings, override: bool = False):
    """Register a new ConversationSettings template."""
    if not override:
        assert template.name not in conv_settings, f"{template.name} has been registered."
    conv_settings[template.name] = template


def get_conv_settings(name: str) -> ConversationSettings:
    """Get a ConversationSettings template."""
    return conv_settings[name].copy()


settings_path = os.path.join(os.path.dirname(__file__), "settings")
for module_loader, name, ispkg in pkgutil.iter_modules([settings_path]):
    # print(module_loader, name, ispkg)
    module = importlib.import_module(".settings." + name, __package__)
    for name, values in vars(module).items():
        if not isinstance(values, ConversationSettings):
            continue
        register_conv_settings(values)






import time
from typing import Any


class CacheState(object):
    def __init__(self, value: Any, ttl: int) -> None:
        self.value = value
        self.ttl = ttl
        self.last_time = time.time()
    
    def set(self, value: Any, ttl: int):
        self.value = value
        self.ttl = ttl
        self.last_time = time.time()
    
    def is_valid(self) -> bool:
        now_time = time.time()
        return now_time - self.last_time < self.ttl
    
    def get(self):
        return self.value


import inspect
import logging

from typing import Any, Callable, Dict, Iterable, Mapping, Optional
from langport.utils.interval_timer import IntervalTimer


class BaseNode(object):
    def __init__(
        self,
        node_addr: str,
        node_id: str,
        logger: logging.Logger,
    ):
        self.node_addr = node_addr
        self.node_id = node_id
        self.logger = logger
        self.online = False

        self.start_fn: Dict[str, Callable[[], None]] = {}
        self.stop_fn: Dict[str, Callable[[], None]] = {}

        self.timers: Dict[str, IntervalTimer] = {}

        self.on_stop("stop_all_timers", self.stop_all_timers)

    async def start(self):
        if self.online:
            return
        for name, fn in self.start_fn.items():
            if inspect.iscoroutinefunction(fn):
                await fn()
            else:
                fn()
        self.online = True

    async def stop(self):
        if not self.online:
            return
        for name, fn in self.stop_fn.items():
            if inspect.iscoroutinefunction(fn):
                await fn()
            else:
                fn()
        self.online = False

    def on_start(self, name: str, fn: Callable[[], None]):
        self.start_fn[name] = fn

    def on_stop(self, name: str, fn: Callable[[], None]):
        self.stop_fn[name] = fn

    def add_timer(
        self,
        name: str,
        interval: float,
        fn: Callable,
        args: Optional[Iterable[Any]] = None,
        kwargs: Optional[Mapping[str, Any]] = None,
        workers: int = 4,
    ) -> bool:
        if name in self.timers:
            return False
        new_timer = IntervalTimer(
            interval=interval, fn=fn, max_workers=workers, args=args, kwargs=kwargs
        )
        self.timers[name] = new_timer
        new_timer.start()
        return True

    def remove_timer(self, name: str) -> bool:
        if name not in self.timers:
            return False
        self.timers[name].cancel()
        del self.timers[name]
        return True

    def stop_all_timers(self):
        for name, timer in self.timers.items():
            timer.cancel()
        self.timers.clear()



from enum import Enum
from typing import Literal, Optional, List, Dict, Any, Set, Union

import time

import shortuuid
from pydantic import BaseModel, Field

# https://code.visualstudio.com/docs/languages/identifiers
class Language(str, Enum):
    UNKNOWN = "unknown"
    PYTHON = "python"
    JAVASCRIPT = "javascript"
    TYPESCRIPT = "typescript"

class EventType(str, Enum):
    COMPLETION = "completion"
    VIEW = "view"
    SELECT = "select"

class Choice(BaseModel):
    index: int
    text: str

class CompletionRequest(BaseModel):
    language: str = "unknown"
    prompt: str

class CompletionResponse(BaseModel):
    id: str = Field(default_factory=lambda: f"modelperm-{shortuuid.random()}")
    created: int = Field(default_factory=lambda: int(time.time()))
    choices: List[Choice] = []


class ChoiceEvent(BaseModel):
    type: Literal["completion", "view", "select"]
    completion_id: str
    choice_index: int


class CompletionEvent(BaseModel):
    type: Literal["completion", "view", "select"]
    id: str = Field(default_factory=lambda: f"modelperm-{shortuuid.random()}")
    language: str = "unknown"
    prompt: str
    created: int = Field(default_factory=lambda: int(time.time()))
    choices: List[Choice] = []

class ValidationError(BaseModel):
    loc: List[Union[int, str]]
    msg: str
    type: str

class HTTPValidationError(BaseModel):
    detail: List[ValidationError] = []

EventTypeMapping = {
    EventType.COMPLETION: CompletionEvent,
    EventType.VIEW: ChoiceEvent,
    EventType.SELECT: ChoiceEvent,
}


'''
Language related
'''


class LanguagePreset(BaseModel):
    max_length: int
    stop_words: List[str]
    reserved_keywords: Optional[Set]


LanguagePresets = {
    Language.UNKNOWN: LanguagePreset(
        max_length=128,
        stop_words=["\n\n"],
    ),
    Language.PYTHON: LanguagePreset(
        max_length=64,
        stop_words=["\n\n", "\ndef", "\n#", "\nimport", "\nfrom", "\nclass"],
        reserved_keywords=set(
            [
                "False",
                "class",
                "from",
                "or",
                "None",
                "continue",
                "global",
                "pass",
                "True",
                "def",
                "if",
                "raise",
                "and",
                "del",
                "import",
                "return",
                "as",
                "elif",
                "in",
                "try",
                "assert",
                "else",
                "is",
                "while",
                "async",
                "except",
                "lambda",
                "with",
                "await",
                "finally",
                "nonlocal",
                "yield",
                "break",
                "for",
                "not",
            ]
        ),
    ),
    Language.JAVASCRIPT: LanguagePreset(
        max_length=128, stop_words=["\n\n", "\nfunction", "\n//", "\nimport", "\nclass"]
    ),
    Language.TYPESCRIPT: LanguagePreset(
        max_length=128,
        stop_words=[
            "\n\n",
            "\nfunction",
            "\n//",
            "\nimport",
            "\nclass",
            "\ninterface",
            "\ntype",
        ],
    ),
}


'''
From huggingface/text-generation-inference
https://github.com/huggingface/text-generation-inference/blob/e7248fe90e27c7c8e39dd4cac5874eb9f96ab182/clients/python/text_generation/types.py
'''

from enum import Enum
from pydantic import BaseModel, validator
from typing import Optional, List

'''
Errors
'''
from typing import Dict


# Text Generation Inference Errors
class ValidationError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class GenerationError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class OverloadedError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class IncompleteGenerationError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


# API Inference Errors
class BadRequestError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class ShardNotReadyError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class ShardTimeoutError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class NotFoundError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class RateLimitExceededError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class NotSupportedError(Exception):
    def __init__(self, model_id: str):
        message = (
            f"Model `{model_id}` is not available for inference with this client. \n"
            "Use `huggingface_hub.inference_api.InferenceApi` instead."
        )
        super(NotSupportedError, self).__init__(message)


# Unknown error
class UnknownError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


def parse_error(status_code: int, payload: Dict[str, str]) -> Exception:
    """
    Parse error given an HTTP status code and a json payload

    Args:
        status_code (`int`):
            HTTP status code
        payload (`Dict[str, str]`):
            Json payload

    Returns:
        Exception: parsed exception

    """
    # Try to parse a Text Generation Inference error
    message = payload["error"]
    if "error_type" in payload:
        error_type = payload["error_type"]
        if error_type == "generation":
            return GenerationError(message)
        if error_type == "incomplete_generation":
            return IncompleteGenerationError(message)
        if error_type == "overloaded":
            return OverloadedError(message)
        if error_type == "validation":
            return ValidationError(message)

    # Try to parse a APIInference error
    if status_code == 400:
        return BadRequestError(message)
    if status_code == 403 or status_code == 424:
        return ShardNotReadyError(message)
    if status_code == 504:
        return ShardTimeoutError(message)
    if status_code == 404:
        return NotFoundError(message)
    if status_code == 429:
        return RateLimitExceededError(message)

    # Fallback to an unknown error
    return UnknownError(message)


'''
Param Models
'''
class Parameters(BaseModel):
    # Activate logits sampling
    do_sample: bool = False
    # Maximum number of generated tokens
    max_new_tokens: int = 20
    # The parameter for repetition penalty. 1.0 means no penalty.
    # See [this paper](https://arxiv.org/pdf/1909.05858.pdf) for more details.
    repetition_penalty: Optional[float] = None
    # Whether to prepend the prompt to the generated text
    return_full_text: bool = False
    # Stop generating tokens if a member of `stop_sequences` is generated
    stop: List[str] = []
    # Random sampling seed
    seed: Optional[int]
    # The value used to module the logits distribution.
    temperature: Optional[float]
    # The number of highest probability vocabulary tokens to keep for top-k-filtering.
    top_k: Optional[int]
    # If set to < 1, only the smallest set of most probable tokens with probabilities that add up to `top_p` or
    # higher are kept for generation.
    top_p: Optional[float]
    # truncate inputs tokens to the given size
    truncate: Optional[int]
    # Typical Decoding mass
    # See [Typical Decoding for Natural Language Generation](https://arxiv.org/abs/2202.00666) for more information
    typical_p: Optional[float]
    # Generate best_of sequences and return the one if the highest token logprobs
    best_of: Optional[int]
    # Watermarking with [A Watermark for Large Language Models](https://arxiv.org/abs/2301.10226)
    watermark: bool = False
    # Get generation details
    details: bool = False

    @validator("best_of")
    def valid_best_of(cls, field_value, values):
        if field_value is not None:
            if field_value <= 0:
                raise ValidationError("`best_of` must be strictly positive")
            if field_value > 1 and values["seed"] is not None:
                raise ValidationError("`seed` must not be set when `best_of` is > 1")
            sampling = (
                values["do_sample"]
                | (values["temperature"] is not None)
                | (values["top_k"] is not None)
                | (values["top_p"] is not None)
                | (values["typical_p"] is not None)
            )
            if field_value > 1 and not sampling:
                raise ValidationError("you must use sampling when `best_of` is > 1")

        return field_value

    @validator("repetition_penalty")
    def valid_repetition_penalty(cls, v):
        if v is not None and v <= 0:
            raise ValidationError("`repetition_penalty` must be strictly positive")
        return v

    @validator("seed")
    def valid_seed(cls, v):
        if v is not None and v < 0:
            raise ValidationError("`seed` must be positive")
        return v

    @validator("temperature")
    def valid_temp(cls, v):
        if v is not None and v <= 0:
            raise ValidationError("`temperature` must be strictly positive")
        return v

    @validator("top_k")
    def valid_top_k(cls, v):
        if v is not None and v <= 0:
            raise ValidationError("`top_k` must be strictly positive")
        return v

    @validator("top_p")
    def valid_top_p(cls, v):
        if v is not None and (v <= 0 or v >= 1.0):
            raise ValidationError("`top_p` must be > 0.0 and < 1.0")
        return v

    @validator("truncate")
    def valid_truncate(cls, v):
        if v is not None and v <= 0:
            raise ValidationError("`truncate` must be strictly positive")
        return v

    @validator("typical_p")
    def valid_typical_p(cls, v):
        if v is not None and (v <= 0 or v >= 1.0):
            raise ValidationError("`typical_p` must be > 0.0 and < 1.0")
        return v


class Request(BaseModel):
    # Prompt
    inputs: str
    # Generation parameters
    parameters: Optional[Parameters]
    # Whether to stream output tokens
    stream: bool = False

    @validator("inputs")
    def valid_input(cls, v):
        if not v:
            raise ValidationError("`inputs` cannot be empty")
        return v

    @validator("stream")
    def valid_best_of_stream(cls, field_value, values):
        parameters = values["parameters"]
        if (
            parameters is not None
            and parameters.best_of is not None
            and parameters.best_of > 1
            and field_value
        ):
            raise ValidationError(
                "`best_of` != 1 is not supported when `stream` == True"
            )
        return field_value


# Prompt tokens
class PrefillToken(BaseModel):
    # Token ID from the model tokenizer
    id: int
    # Token text
    text: str
    # Logprob
    # Optional since the logprob of the first token cannot be computed
    logprob: Optional[float]


# Generated tokens
class Token(BaseModel):
    # Token ID from the model tokenizer
    id: int
    # Token text
    text: str
    # Logprob
    logprob: float
    # Is the token a special token
    # Can be used to ignore tokens when concatenating
    special: bool


# Generation finish reason
class FinishReason(str, Enum):
    # number of generated tokens == `max_new_tokens`
    Length = "length"
    # the model generated its end of sequence token
    EndOfSequenceToken = "eos_token"
    # the model generated a text included in `stop_sequences`
    StopSequence = "stop_sequence"


# Additional sequences when using the `best_of` parameter
class BestOfSequence(BaseModel):
    # Generated text
    generated_text: str
    # Generation finish reason
    finish_reason: FinishReason
    # Number of generated tokens
    generated_tokens: int
    # Sampling seed if sampling was activated
    seed: Optional[int]
    # Prompt tokens
    prefill: List[PrefillToken]
    # Generated tokens
    tokens: List[Token]


# `generate` details
class Details(BaseModel):
    # Generation finish reason
    finish_reason: FinishReason
    # Number of generated tokens
    generated_tokens: int
    # Sampling seed if sampling was activated
    seed: Optional[int]
    # Prompt tokens
    prefill: List[PrefillToken]
    # Generated tokens
    tokens: List[Token]
    # Additional sequences when using the `best_of` parameter
    best_of_sequences: Optional[List[BestOfSequence]]


# `generate` return value
class Response(BaseModel):
    # Generated text
    generated_text: str
    # Generation details
    details: Details


# `generate_stream` details
class StreamDetails(BaseModel):
    # Generation finish reason
    finish_reason: FinishReason
    # Number of generated tokens
    generated_tokens: int
    # Sampling seed if sampling was activated
    seed: Optional[int]


# `generate_stream` return value
class StreamResponse(BaseModel):
    # Generated token
    token: Token
    # Complete generated text
    # Only available when the generation is finished
    generated_text: Optional[str]
    # Generation details
    # Only available when the generation is finished
    details: Optional[StreamDetails]


# Inference API currently deployed model
class DeployedModel(BaseModel):
    model_id: str
    sha: str



from concurrent.futures import ThreadPoolExecutor
import inspect
import threading
import time
import asyncio
import traceback
from typing import Any, Iterable, Mapping, Optional


class IntervalTimer(object):
    def __init__(
        self,
        interval: float,
        fn: object,
        max_workers: int = 4,
        args: Optional[Iterable[Any]] = None,
        kwargs: Optional[Mapping[str, Any]] = None,
    ) -> None:
        self.interval = interval
        self.fn = fn
        self.max_workers = max_workers
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}

        self._timer = threading.Thread(target=self.run)
        self._executor = ThreadPoolExecutor(max_workers=max_workers)
        self.is_activated = False
        self.last_time = time.time()

    def start(self):
        self.is_activated = True
        self.last_time = time.time()
        self._timer.start()

    def cancel(self):
        self.is_activated = False
        self._timer.join()

    def function_wrapper(
        self,
        args: Optional[Iterable[Any]] = None,
        kwargs: Optional[Mapping[str, Any]] = None,
    ):
        try:
            if inspect.iscoroutinefunction(self.fn):
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                loop.run_until_complete(self.fn(*args, **kwargs))
                loop.close()
            else:
                self.fn(*args, **kwargs)
        except:
            traceback.print_exc()

    def run(self):
        while True:
            if self.is_activated:
                start_time = time.time()
                if start_time - self.last_time > self.interval:
                    self._executor.submit(self.function_wrapper, self.args, self.kwargs)
                    self.last_time = time.time()
                    # Never run this code, because finish_time - start_time = 0.
                    # if finish_time - start_time > self.interval:
                    #     print(
                    #         f"Overloaded!! Last timer using {finish_time - start_time}s."
                    #     )
                    # else:
                        # time.sleep(0.01)
                else:
                    time.sleep(0.01)
            else:
                break



from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
)

from langport.data.conversation import ConversationHistory, SeparatorStyle
from langport.data.conversation.conversation_settings import get_conv_settings
from langport.model.model_adapter import BaseAdapter


class StableLMAdapter(BaseAdapter):
    """The model adapter for StabilityAI/stablelm-tuned-alpha-7b"""

    def match(self, model_path: str):
        return "stablelm" in model_path

    def get_default_conv_template(self, model_path: str) -> ConversationHistory:
        settings = get_conv_settings("stablelm")
        return ConversationHistory(
            system="""<|SYSTEM|># StableLM Tuned (Alpha version)
- StableLM is a helpful and harmless open-source AI language model developed by StabilityAI.
- StableLM is excited to be able to help the user, but will refuse to do anything that could be considered harmful to the user.
- StableLM is more than just an information source, StableLM is also able to write poetry, short stories, and make jokes.
- StableLM will refuse to participate in anything that could harm a human.
""",
            messages=(),
            offset=0,
            settings=settings,
        )



import json
from typing import List

from langport.core.cluster_worker import ClusterWorker
from langport.model.executor.generation import GenerationExecutor

from langport.protocol.worker_protocol import (
    BaseWorkerResult,
    GenerationTask,
    GenerationWorkerResult,
    UsageInfo,
)

from langport.constants import (
    GENERATION_INFERENCE_INTERVAL,
    WORKER_API_TIMEOUT,
    WORKER_HEART_BEAT_INTERVAL,
    ErrorCode,
)
from langport.utils import server_error_msg, pretty_print_semaphore


class GenerationModelWorker(ClusterWorker):
    def __init__(
        self,
        node_addr: str,
        node_id: str,
        init_neighborhoods_addr: List[str],
        executor: GenerationExecutor,
        limit_model_concurrency: int,
        max_batch: int,
        stream_interval: int,
        logger
    ):
        super(GenerationModelWorker, self).__init__(
            node_addr=node_addr,
            node_id=node_id,
            init_neighborhoods_addr=init_neighborhoods_addr,
            limit_model_concurrency=limit_model_concurrency,
            max_batch=max_batch,
            stream_interval=stream_interval,
            logger=logger,
        )
        self.executor = executor
        workers = max(1, 2 * self.limit_model_concurrency // self.max_batch)
        self.add_timer(
            "generation_inference",
            GENERATION_INFERENCE_INTERVAL,
            self.executor.inference,
            args=[self,],
            kwargs=None,
            workers=workers,
        )
    
        self.on_start("set_features", self.set_features)
        self.on_start("set_model_name", self.set_model_name)

    async def set_features(self):
        await self.set_local_state("features", ["generation"], ttl=360)
    
    async def set_model_name(self):
        await self.set_local_state("model_name", self.executor.model_name, ttl=360)


    async def generation_stream(self, task: GenerationTask):
        prompt_tokens = len(self.executor.tokenize(task.prompt))
        max_tokens = task.max_tokens
        context_length = self.executor.context_length

        if context_length is not None and prompt_tokens + max_tokens > context_length:
            yield BaseWorkerResult(task_id=task.task_id,
                                   type="error",
                                   message=f"This model's maximum context length is {context_length} tokens. "
                                    f"However, you requested {max_tokens + prompt_tokens} tokens "
                                    f"({prompt_tokens} in the messages, "
                                    f"{max_tokens} in the completion). "
                                    f"Please reduce the length of the messages or completion.",
                                    error_code=ErrorCode.CONTEXT_OVERFLOW
                                   )
            return

        await self.add_task(task)
        async for chunk in self.fetch_task_result(task.task_id):
            yield chunk

    async def generation_bytes_stream(self, task: GenerationTask):
        async for chunk in self.generation_stream(task):
            yield json.dumps(chunk.dict()).encode() + b"\0"



import os
from types import SimpleNamespace
import warnings

import torch

os.environ["RWKV_JIT_ON"] = "1"
os.environ["RWKV_CUDA_ON"] = "1"

from rwkv.model import RWKV
from rwkv.utils import PIPELINE, PIPELINE_ARGS


class RwkvModel:
    def __init__(self, model_path):
        warnings.warn(
            "Experimental support. Please use ChatRWKV if you want to chat with RWKV"
        )
        self.config = SimpleNamespace(is_encoder_decoder=False)
        self.model = RWKV(model=model_path, strategy="cuda fp16")

    def to(self, target):
        assert target == "cuda"

    def __call__(self, input_ids, use_cache, past_key_values=None):
        assert use_cache == True
        input_ids = input_ids[0].detach().cpu().numpy()
        # print(input_ids)
        logits, state = self.model.forward(input_ids, past_key_values)
        # print(logits)
        logits = logits.unsqueeze(0).unsqueeze(0)
        out = SimpleNamespace(logits=logits, past_key_values=state)
        return out



import traceback
from typing import List, Optional

import torch
from langport.model.executor.huggingface import HuggingfaceExecutor
from langport.protocol.worker_protocol import BaseWorkerResult, EmbeddingWorkerResult, UsageInfo
from langport.workers.embedding_worker import EmbeddingModelWorker


class HuggingfaceEmbeddingExecutor(HuggingfaceExecutor):
    def __init__(
        self,
        model_name: str,
        model_path: str,
        device: str,
        num_gpus: int,
        max_gpu_memory: Optional[str],
        load_8bit: bool,
        cpu_offloading: bool,
        deepspeed: bool = False,
        trust_remote_code: bool = False,
    ) -> None:
        super(HuggingfaceEmbeddingExecutor, self).__init__(
            model_name=model_name,
            model_path=model_path,
            device=device,
            num_gpus=num_gpus,
            max_gpu_memory=max_gpu_memory,
            load_8bit=load_8bit,
            cpu_offloading=cpu_offloading
        )
        self.adapter = None
        self.model = None
        self.tokenizer = None
        self.adapter, self.model, self.tokenizer = self.load_model(
            model_path, device, num_gpus, max_gpu_memory, load_8bit, cpu_offloading, deepspeed, trust_remote_code
        )

        if hasattr(self.model.config, "max_sequence_length"):
            self._context_len = self.model.config.max_sequence_length
        elif hasattr(self.model.config, "max_position_embeddings"):
            self._context_len = self.model.config.max_position_embeddings
        else:
            self._context_len = 2048
    
    @property
    def context_length(self) -> int:
        return self._context_len

    def tokenize(self, text: str) -> List[int]:
        input_ids = self.tokenizer(text).input_ids
        return input_ids
    
    # Mean Pooling - Take attention mask into account for correct averaging
    def _mean_pooling(self, model_output, attention_mask):
        token_embeddings = model_output[0]  # First element of model_output contains all token embeddings
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)
    
    @torch.inference_mode()
    def inference(self, worker: "EmbeddingModelWorker"):
        if not worker.online:
            return
        tasks = worker.fetch_tasks()
        batch_size = len(tasks)
        if batch_size == 0:
            return
        # print(batch_size)

        prompts = [task.input for task in tasks]
        try:
            tokenizer = self.tokenizer
            model = self.model

            # ValueError: Asking to pad but the tokenizer does not have a padding token.
            # Please select a token to use as `pad_token` `(tokenizer.pad_token = tokenizer.eos_token e.g.)`
            # or add a new pad token via `tokenizer.add_special_tokens({'pad_token': '[PAD]'})`.
            if tokenizer._pad_token is None:
                tokenizer.pad_token = tokenizer.eos_token

            encoded_prompts = tokenizer(prompts, return_tensors="pt", padding="longest").to(self.device)
            input_ids = encoded_prompts.input_ids
            if model.config.is_encoder_decoder:
                decoder_input_ids = torch.full(
                    (batch_size, 1),
                    model.generation_config.decoder_start_token_id,
                    dtype=torch.long,
                    device=self.device,
                )
                model_output = model(input_ids, decoder_input_ids=decoder_input_ids, output_hidden_states=True)
                data = model_output.decoder_hidden_states[-1]
            elif model.config.is_decoder:
                model_output = model(input_ids, output_hidden_states=True)
                is_chatglm = "chatglm" in str(type(model)).lower()
                if is_chatglm:
                    data = model_output.hidden_states[-1].transpose(0, 1)
                else:
                    data = model_output.hidden_states[-1]
            else:
                data = model(**encoded_prompts)
            # embeddings = torch.mean(data, dim=1)
            embeddings = self._mean_pooling(data, encoded_prompts['attention_mask'])
            for i in range(batch_size):
                token_num = len(tokenizer(prompts[i]).input_ids)
                worker.push_task_result(
                    tasks[i].task_id,
                    EmbeddingWorkerResult(
                        task_id=tasks[i].task_id,
                        type="data",
                        embedding=embeddings[i].tolist(),
                        usage=UsageInfo(prompt_tokens=token_num, total_tokens=token_num),
                    )
                )

        except torch.cuda.OutOfMemoryError:
            for i in range(batch_size):
                worker.push_task_result(
                    tasks[i].task_id,
                    BaseWorkerResult(
                        task_id=tasks[i].task_id,
                        type="error",
                        message="Cuda out of Memory Error"
                    )
                )
        except Exception as e:
            traceback.print_exc()
            for i in range(batch_size):
                worker.push_task_result(
                    tasks[i].task_id,
                    BaseWorkerResult(
                        task_id=tasks[i].task_id,
                        type="error",
                        message=str(e)
                    )
                )
        
        for i in range(batch_size):
            worker.push_task_result(
                tasks[i].task_id,
                BaseWorkerResult(
                    task_id=tasks[i].task_id,
                    type="done",
                )
            )



import asyncio
from collections import defaultdict
import logging
import json
import re
from typing import Dict, List, Optional

import queue
from langport.core.cluster_node import ClusterNode

from langport.protocol.worker_protocol import (
    BaseWorkerResult,
    BaseWorkerTask,
    WorkerAddressRequest,
    WorkerAddressResponse,
)

from langport.constants import (
    HEART_BEAT_EXPIRATION,
    WORKER_API_TIMEOUT,
    WORKER_HEART_BEAT_INTERVAL,
    ErrorCode,
)
from langport.utils.evaluation import safe_eval

class ClusterWorker(ClusterNode):
    def __init__(
        self,
        node_addr: str,
        node_id: str,
        init_neighborhoods_addr: List[str],
        limit_model_concurrency: int,
        max_batch: int,
        stream_interval: int,
        logger: logging.Logger,
    ):
        super(ClusterWorker, self).__init__(
            node_addr=node_addr,
            node_id=node_id,
            init_neighborhoods_addr=init_neighborhoods_addr,
            logger=logger,
        )

        self.limit_model_concurrency = limit_model_concurrency
        self.max_batch = max_batch
        self.stream_interval = stream_interval

        self.global_counter = 0
        self.model_semaphore = None

        self.task_queue: queue.Queue[BaseWorkerTask] = queue.Queue()
        self.output_queue: Dict[str, queue.Queue[BaseWorkerResult]] = defaultdict(
            queue.Queue
        )

        self.on_start("set_queue_state", self.set_queue_state)
        self.on_start("set_features", self.set_features)
        self.on_start("set_speed", self.set_speed)

    async def set_queue_state(self):
        await self.set_local_state("queue_length", self.get_queue_length(), ttl=16)

    async def set_features(self):
        await self.set_local_state("features", [], ttl=360)
    
    async def set_speed(self):
        await self.set_local_state("speed", 1.0, ttl=360)

    def release_model_semaphore(self):
        self.model_semaphore.release()

    def acquire_model_semaphore(self):
        self.global_counter += 1
        if self.model_semaphore is None:
            self.model_semaphore = asyncio.Semaphore(self.limit_model_concurrency)
        return self.model_semaphore.acquire()

    def get_queue_length(self):
        if (
            self.model_semaphore is None
            or self.model_semaphore._value is None
            or self.model_semaphore._waiters is None
        ):
            return 0
        else:
            return (
                self.task_queue.qsize()
                + len(self.model_semaphore._waiters)
            )

    async def add_task(self, task: BaseWorkerTask):
        self.task_queue.put(task, block=True, timeout=WORKER_API_TIMEOUT)
        self.output_queue[task.task_id] = queue.Queue()
        await self.set_queue_state()

    async def fetch_task_result(self, task_id: str):
        result_queue = self.output_queue[task_id]
        retry_counter = 0
        while True:
            await self.set_queue_state()
            try:
                event = result_queue.get(block=False, timeout=None)
            except queue.Empty:
                await asyncio.sleep(0.01)
                retry_counter += 1
                # If client disconnected, stop to wait queue.
                if retry_counter > 2000:
                    break
                else:
                    continue
            retry_counter = 0
            if event.type == "done":
                break
            elif event.type == "error":
                yield event
                break
            elif event.type == "finish":
                yield event
                break
            elif event.type == "data":
                yield event
            else:
                raise ValueError("Bad chunk type.")

        del self.output_queue[task_id]

    def get_num_tasks(self) -> int:
        return self.task_queue.qsize()

    def fetch_tasks(self) -> List[BaseWorkerResult]:
        task_batch = []
        while len(task_batch) < self.max_batch:
            try:
                task = self.task_queue.get(block=False, timeout=None)
            except queue.Empty:
                break
            task_batch.append(task)
        return task_batch

    def push_task_result(self, task_id: str, response: BaseWorkerResult):
        self.output_queue[task_id].put(response, block=True, timeout=WORKER_API_TIMEOUT)

    async def api_get_worker_address(self, request: WorkerAddressRequest) -> WorkerAddressResponse:
        id_list, address_list, values = await self.get_worker_address(request.condition, request.expression)
        return WorkerAddressResponse(
            id_list=id_list,
            address_list=address_list,
            values=values,
        )
    
    async def get_worker_address(self, condition: str, expression: str) -> Optional[str]:
        condition_variables = re.findall(r'\{(.*?)\}', condition)
        expression_variables = re.findall(r'\{(.*?)\}', expression)

        worker_ids = []
        worker_address = []
        worker_values = []
        for w_id, w_info in self.neighborhoods.items():
            final_condition = condition
            final_condition_variables = {}
            for v_name in condition_variables:
                variable_value = await self.get_node_state(w_info.node_id, v_name)
                final_condition = final_condition.replace("{" + v_name + "}", v_name)
                final_condition_variables[v_name] = variable_value
            
            final_expression = expression
            final_expression_variables = {}
            for v_name in expression_variables:
                variable_value = await self.get_node_state(w_info.node_id, v_name)
                final_expression = final_expression.replace("{" + v_name + "}", v_name)
                final_expression_variables[v_name] = variable_value

            if safe_eval(final_condition, final_condition_variables):
                worker_ids.append(w_id)
                worker_address.append(w_info.node_addr)
                value_json = json.dumps(safe_eval(final_expression, final_expression_variables))
                worker_values.append(value_json)
        
        return worker_ids, worker_address, worker_values
                


from langport.data.conversation import (
    ConversationSettings,
    SeparatorStyle,
)


billa = ConversationSettings(
    name="billa",
    roles=("Human", "Assistant"),
    sep_style=SeparatorStyle.ADD_COLON_SPACE_SINGLE,
    sep="\n",
    stop_str="Human:",
)


from langport.data.conversation import (
    ConversationSettings,
    SeparatorStyle,
)


# Koala default template
koala_v1 = ConversationSettings(
    name="koala_v1",
    roles=("USER", "GPT"),
    sep_style=SeparatorStyle.ADD_COLON_TWO,
    sep=" ",
    sep2="</s>",
)



from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
)

from langport.data.conversation import ConversationHistory, SeparatorStyle
from langport.data.conversation.conversation_settings import get_conv_settings
from langport.model.model_adapter import BaseAdapter

class OasstPythiaAdapter(BaseAdapter):
    """The model adapter for OpenAssistant/oasst-sft-1-pythia-12b"""

    def match(self, model_path: str):
        return "oasst" in model_path and "pythia" in model_path
    
    def get_default_conv_template(self, model_path: str) -> ConversationHistory:
        settings = get_conv_settings("oasst_pythia")
        return ConversationHistory(
            system="BEGINNING OF CONVERSATION:",
            messages=(),
            offset=0,
            settings=settings,
        )



from langport.data.conversation import (
    ConversationSettings,
    SeparatorStyle,
)

# Buddy default template
openbuddy = ConversationSettings(
    name="openbuddy",
    roles=("User", "Assistant"),
    sep_style=SeparatorStyle.ADD_COLON_SINGLE,
    sep="\n",
)



import asyncio
from collections import defaultdict
import logging
import json
import time
from typing import Any, Dict, List

import httpx
from langport.core.base_node import BaseNode

from langport.protocol.worker_protocol import (
    GetNodeStateRequest,
    GetNodeStateResponse,
    NodeInfo,
    NodeInfoRequest,
    NodeInfoResponse,
    NodeListRequest,
    NodeListResponse,
    RegisterNodeRequest,
    RegisterNodeResponse,
    RemoveNodeRequest,
    RemoveNodeResponse,
    HeartbeatPing,
    HeartbeatPong,
)

from langport.constants import (
    HEART_BEAT_EXPIRATION,
    WORKER_API_TIMEOUT,
    WORKER_HEART_BEAT_INTERVAL,
    ErrorCode,
)
from langport.utils.cache_state import CacheState
from langport.utils.http_pool import AsyncHttpPool
from cachetools import LRUCache, TTLCache
from asyncache import cached

class ClusterNode(BaseNode):
    def __init__(
        self,
        node_addr: str,
        node_id: str,
        init_neighborhoods_addr: List[str],
        logger: logging.Logger,
    ):
        super(ClusterNode, self).__init__(
            node_addr=node_addr,
            node_id=node_id,
            logger=logger,
        )
        self.init_neighborhoods_addr: List[str] = init_neighborhoods_addr
        self.neighborhoods: Dict[str, NodeInfo] = {}
        self.headers = {"User-Agent": "LangPort nodes"}

        self.states: Dict[str, CacheState] = {}
        self.remote_states: Dict[str, Dict[str, CacheState]] = defaultdict(dict)

        self.client = AsyncHttpPool()
        
        # timers
        self.add_timer(
            "heartbeat",
            WORKER_HEART_BEAT_INTERVAL,
            self.send_heartbeat_broadcast,
            args=None,
            kwargs=None,
            workers=1,
        )

        self.add_timer(
            "expiration_check",
            HEART_BEAT_EXPIRATION // 2,
            self.remove_nodes_expiration,
            args=None,
            kwargs=None,
            workers=1,
        )

        # start and stop
        self.on_start("get_all_init_neighborhoods", self.get_all_init_neighborhoods)
        self.on_start("register_node_broadcast", self.register_self_node_broadcast)
        self.on_stop("remove_node_broadcast", self.remove_self_node_broadcast)
    
    def _add_node(self, node_id: str, node_addr: str, check_heart_beat: bool=True):
        self.neighborhoods[node_id] = NodeInfo(
            node_id=node_id,
            node_addr=node_addr,
            check_heart_beat=check_heart_beat,
            refresh_time=int(time.time())
        )
    
    def _update_node(self, node_id: str, node_addr: str, check_heart_beat: bool=True):
        self.neighborhoods[node_id] = NodeInfo(
            node_id=node_id,
            node_addr=node_addr,
            check_heart_beat=check_heart_beat,
            refresh_time=int(time.time())
        )
    
    def _remove_node(self, node_id: str):
        if node_id in self.neighborhoods:
            del self.neighborhoods[node_id]
    
    async def get_all_init_neighborhoods(self):
        for neighbor_addr in self.init_neighborhoods_addr:
            try:
                info = await self.get_node_info(neighbor_addr)
            except httpx.ReadTimeout:
                self.logger.error(f"Cannot connect to neighbor {neighbor_addr}.")
                continue
            self._add_node(info.node_id, info.node_addr, info.check_heart_beat)
        # add self
        self._add_node(self.node_id, self.node_addr, True)
    
    async def get_node_info(self, node_addr: str) -> NodeInfo:
        response = await self.client.post(
            node_addr + "/node_info",
            headers=self.headers,
            json=NodeInfoRequest(node_id=self.node_id).dict(),
            timeout=WORKER_API_TIMEOUT,
        )
        remote_node_info = NodeInfoResponse.parse_obj(response.json())
        return remote_node_info.node_info

    async def register_node(self, target_node_addr: str, register_node_id: str, register_node_addr: str) -> bool:
        self.logger.info(f"Register {register_node_addr} to node {target_node_addr}")

        if target_node_addr == self.node_addr:
            self._add_node(register_node_id, register_node_addr)
            return True
        else:
            data = RegisterNodeRequest(
                node_id=register_node_id,
                node_addr=register_node_addr,
                check_heart_beat=True,
            )
        
            response = await self.client.post(
                target_node_addr + "/register_node",
                headers=self.headers,
                json=data.dict(),
                timeout=WORKER_API_TIMEOUT,
            )
            remote = RegisterNodeResponse.parse_obj(response.json())
            self._add_node(remote.node_id, remote.node_addr)
            
            # fetch remote node info
            remote_nodes = await self.fetch_all_nodes(remote.node_addr)
            for node in remote_nodes.nodes:
                self._update_node(node_id=node.node_id, node_addr=node.node_addr)
            return True

    async def register_node_broadcast(self, register_node_id: str, register_node_addr: str):
        self.logger.info(f"Register node broadcast. node_id: {register_node_id}, node_addr: {register_node_addr}")

        neighborhoods = [(k, v) for k, v in self.neighborhoods.items()]
        for node_id, node_info in neighborhoods:
            if node_info.node_addr == register_node_addr:
                continue
            await self.register_node(node_info.node_addr, register_node_id, register_node_addr)
    
    async def register_self_node_broadcast(self):
        await self.register_node_broadcast(self.node_id, self.node_addr)

    async def remove_node(self, target_node_addr: str, removed_node_id: str) -> bool:
        self.logger.info(f"Remove node {removed_node_id} from {target_node_addr}")

        if target_node_addr == self.node_addr:
            return True

        data = RemoveNodeRequest(
            node_id=removed_node_id,
        )

        response = await self.client.post(
            target_node_addr + "/remove_node",
            headers=self.headers,
            json=data.dict(),
            timeout=WORKER_API_TIMEOUT,
        )
        ret = RemoveNodeResponse.parse_obj(response.json())

        return True

    async def remove_node_broadcast(self, removed_node_id: str):
        self.logger.info(f"Remove node broadcast. node_addr: {removed_node_id}")
        neighborhoods = [(k, v) for k, v in self.neighborhoods.items()]
        for node_id, node_info in neighborhoods:
            if node_id == removed_node_id:
                continue
            await self.remove_node(node_info.node_addr, removed_node_id)

    async def remove_self_node_broadcast(self):
        await self.remove_node_broadcast(self.node_id)

    async def send_heartbeat(self, node_addr: str):
        data = HeartbeatPing(
            node_id=self.node_id,
        )
        response = await self.client.post(
            node_addr + "/heartbeat",
            headers=self.headers,
            json=data.dict(),
            timeout=WORKER_API_TIMEOUT,
        )
        ret = HeartbeatPong.parse_obj(response.json())
  
        return ret

    async def send_heartbeat_broadcast(self):
        self.logger.info(
            f"Send heartbeat. Worker: {self.node_id}; Address: {self.node_addr}."
        )

        self.logger.info(
            f"Neighborhoods: {[i for i, v in self.neighborhoods.items()]}."
        )

        for node_id, node_info in self.neighborhoods.items():
            if node_id == self.node_id:
                continue
            await self.send_heartbeat(node_info.node_addr)
        
    async def fetch_all_nodes(self, node_addr: str):
        data = NodeListRequest(
            node_id=self.node_id,
        )

        response = await self.client.post(
            node_addr + "/node_list",
            headers=self.headers,
            json=data.dict(),
            timeout=WORKER_API_TIMEOUT,
        )
        ret = NodeListResponse.parse_obj(response.json())
  
        return ret

    def remove_nodes_expiration(self):
        neighborhoods = [(k, v) for k, v in self.neighborhoods.items()]
        for node_id, node_info in neighborhoods:
            if node_id == self.node_id:
                continue
            if time.time() - node_info.refresh_time > HEART_BEAT_EXPIRATION:
                self.logger.info(f"Node {node_id} is expired.")
                self._remove_node(node_id)

    @cached(TTLCache(maxsize=1024, ttl=16))
    async def request_node_state(self, node_addr: str, name: str) -> GetNodeStateResponse:
        data = GetNodeStateRequest(
            state_name=name,
        )

        response = await self.client.post(
            node_addr + "/get_node_state",
            headers=self.headers,
            json=data.dict(),
            timeout=WORKER_API_TIMEOUT,
        )
        ret = GetNodeStateResponse.parse_obj(response.json())

        return ret
    
    async def _get_node_state_local(self, node_id: str, name: str) -> Any:
        entry = self.states.get(name, None)
        if entry is not None:
            return entry.get()
        else:
            return None
    
    @cached(TTLCache(maxsize=1024, ttl=1))
    async def _get_node_state_remote(self, node_id: str, name: str) -> Any:
        target_state = self.remote_states[node_id]
        if name in target_state:
            if target_state[name].is_valid():
                return target_state[name].get()

        node_info = self.neighborhoods[node_id]
        response = await self.request_node_state(node_info.node_addr, name)
        value = json.loads(response.state_value)
        ttl = response.state_ttl
        if value is not None:
            # print(f"refresh state {name}, node id {node_id}, value {value}")
            self.remote_states[node_id][name] = CacheState(value, ttl)
            return self.remote_states[node_id][name].get()
        else:
            return None
    
    async def get_node_state(self, node_id: str, name: str) -> Any:
        if node_id == self.node_id:
            return await self._get_node_state_local(node_id, name)
        else:
            state_lock = asyncio.Lock()
            async with state_lock:
                ret = await self._get_node_state_remote(node_id, name)
            return ret
    
    async def set_local_state(self, name: str, value: Any, ttl: int=60):
        self.states[name] = CacheState(value, ttl)

    async def api_register_node(self, request: RegisterNodeRequest) -> RegisterNodeResponse:
        if request.node_id not in self.neighborhoods:
            self.logger.info(f"Register {request.node_id} on {self.node_id}")
            self._add_node(request.node_id, request.node_addr)

            # broacast again
            await self.register_node_broadcast(request.node_id, request.node_addr)
        
        return RegisterNodeResponse(
            node_id=self.node_id,
            node_addr=self.node_addr,
            check_heart_beat=True
        )

    async def api_remove_node(self, request: RemoveNodeRequest) -> RemoveNodeResponse:
        if request.node_id in self.neighborhoods:
            self.logger.info(f"Remove {request.node_id} from {self.node_id}")
            self._remove_node(request.node_id)

            # broacast again
            await self.remove_node_broadcast(request.node_id)
        return RemoveNodeResponse(node_id=self.node_id)
    
    async def api_receive_heartbeat(self, request: HeartbeatPing) -> HeartbeatPong:
        if request.node_id in self.neighborhoods:
            self.neighborhoods[request.node_id].refresh_time = int(time.time())
        else:
            self.logger.info(f"Invalid ping packet from {request.node_id}.")
        return HeartbeatPong(exist=True)
    
    async def api_return_node_list(self, request: NodeListRequest) -> NodeListResponse:
        node_list = [node_info for node_id, node_info in self.neighborhoods.items()]
        return NodeListResponse(nodes=node_list)
    
    async def api_return_node_info(self, request: NodeInfoRequest) -> NodeInfoResponse:
        return NodeInfoResponse(
            node_info=NodeInfo(
                node_id=self.node_id,
                node_addr=self.node_addr,
                check_heart_beat=True,
            )
        )
    
    async def api_return_node_state(self, request: GetNodeStateRequest) -> GetNodeStateResponse:
        if request.state_name in self.states:
            entry = self.states[request.state_name]
            return GetNodeStateResponse(
                state_value=json.dumps(entry.value),
                state_ttl=entry.ttl
            )
        else:
            return GetNodeStateResponse(
                state_value=json.dumps(None)
            )


from langport.data.conversation import (
    ConversationSettings,
    SeparatorStyle,
)


one_shot = ConversationSettings(
    name="one_shot",
    roles=("Human", "Assistant"),
    sep_style=SeparatorStyle.ADD_COLON_SINGLE,
    sep="\n### ",
    stop_str="###",
)



import argparse
import asyncio
import json
import logging
import httpx
import pandas as pd
import graphyte

from langport.protocol.worker_protocol import WorkerAddressRequest, WorkerAddressResponse
from langport.routers.gateway.common import AppSettings


logger = logging.getLogger(__name__)

async def list_workers(app_settings: AppSettings):
    data_dict = {
        "address": [],
        "features": [],
        "model": [],
        "queue_length": [],
    }
    async with httpx.AsyncClient() as client:
        payload = WorkerAddressRequest(
            condition="True", expression="{model_name}, {features}, {queue_length}"
        )
        ret = await client.post(
            app_settings.controller_address + "/get_worker_address",
            json=payload.dict(),
        )
        if ret.status_code != 200:
            return []
        response = WorkerAddressResponse.parse_obj(ret.json())
        address_list = response.address_list
        data = [json.loads(obj) for obj in response.values]

    for i in range(len(address_list)):
        data_dict["address"].append(address_list[i])
        data_dict["features"].append(str(data[i][1]))
        data_dict["model"].append(data[i][0])
        data_dict["queue_length"].append(data[i][2])
    
    return pd.DataFrame(data_dict)

async def main(app_settings):
    graphyte.init(args.graphite_address, prefix=args.prefix)

    while True:
        df = await list_workers(app_settings)
        for i, row in df.iterrows():
            address_name = row['address'].replace('.', '_')
            graphyte.send(f"langport.{address_name}.queue_length", row["queue_length"])

        await asyncio.sleep(3)


if __name__ in ["__main__", "langport.service.gateway.graphite_feeder"]:
    parser = argparse.ArgumentParser(
        description="Langport Cluster Graphite Data Monitor."
    )
    parser.add_argument("--host", type=str, default="localhost", help="host name")
    parser.add_argument("--port", type=int, default=8000, help="port number")
    parser.add_argument(
        "--controller-address", type=str, default="http://localhost:21001"
    )
    parser.add_argument(
        "--graphite-address", type=str, default="graphite.example.com"
    )
    parser.add_argument(
        "--prefix", type=str, default="system.sync"
    )
    args = parser.parse_args()

    logger.debug(f"==== args ====\n{args}")

    app_settings = AppSettings(controller_address=args.controller_address)
    
    asyncio.run(main(app_settings))